#
# Copyright (c) 2017 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""
Module for abstract base class for checkpoint object and checkpoint collection
"""
from typing import Any, Dict, List


class Saver(object):
    """
    ABC for saver objects that implement saving/restoring to/from path, and merging two savers.
    """
    @property
    def path(self):
        """
        Relative path for save/load. If two saver objects return the same path, they must be merge-able.
        """
        raise NotImplementedError

    def save(self, sess: Any, save_path: str) -> List[str]:
        """
        Save to save_path
        :param sess: active session for session-based frameworks (e.g. TF)
        :param save_path: full path to save checkpoint (typically directory plus self.path plus checkpoint count).
        :return: list of all saved paths
        """
        raise NotImplementedError

    def restore(self, sess: Any, restore_path: str) -> None:
        """
        Restore from restore_path
        :param sess: active session for session-based frameworks (e.g. TF)
        :param restore_path: full path to load checkpoint from.
        """
        raise NotImplementedError

    def merge(self, other: 'Saver') -> None:
        """
        Merge other saver into this saver
        :param other: saver to be merged into self
        """
        raise NotImplementedError


class SaverCollection(object):
    """
    Object for storing a collection of saver objects. It takes care of ensuring uniqueness of saver paths
    and merging savers if they have the same path. For example, if a saver handles saving a generic key/value
    file for all networks in a single file, it can use a more generic path and all savers of all networks would be
    merged into a single saver that saves/restores parameters for all networks.
    NOTE: If two savers have the same path, the respective saver class must support merging them
    into a single saver that saves/restores all merged parameters.
    """
    def __init__(self, saver: Saver = None):
        """
        :param saver: optional initial saver for the collection
        """
        self._saver_dict = dict()  # type: Dict[str, Saver]
        if saver is not None:
            self._saver_dict[saver.path] = saver

    def add(self, saver: Saver):
        """
        Add a new saver to the collection. If saver.path is already in the collection, merge
        the new saver with the existing saver.
        :param saver: new saver to be added to collection
        """
        if saver.path in self._saver_dict:
            self._saver_dict[saver.path].merge(saver)
        else:
            self._saver_dict[saver.path] = saver

    def update(self, other: 'SaverCollection'):
        """
        Merge savers from other collection into self
        :param other: saver collection to update self with.
        """
        for c in other:
            self.add(c)

    def save(self, sess: Any, save_path: str) -> List[str]:
        """
        Call save on all savers in the collection
        :param sess: active session for session-based frameworks (e.g. TF)
        :param save_path: path for saving checkpoints using savers. All saved file paths must
        start with this path in their full path. For example if save_path is '/home/checkpoints/checkpoint-01',
        then saved file paths can be '/home/checkpoints/checkpoint-01.main-network' but not
        '/home/checkpoints/main-network'
        :return: list of all saved paths
        """
        paths = list()
        for saver in self:
            paths.extend(saver.save(sess, self._full_path(save_path, saver)))
        return paths

    def restore(self, sess: Any, restore_path: str) -> None:
        """
        Call restore on all savers in the collection
        :param sess: active session for session-based frameworks (e.g. TF)
        :param restore_path: path for restoring checkpoint using savers.
        """
        for saver in self:
            saver.restore(sess, self._full_path(restore_path, saver))

    def __iter__(self):
        """
        Return an iterator for savers in the collection
        :return: saver iterator
        """
        return (v for v in self._saver_dict.values())

    @staticmethod
    def _full_path(path_prefix: str, saver: Saver) -> str:
        """
        Concatenates path of the saver to parent prefix to create full save path
        :param path_prefix: prefix of the path
        :param saver: saver object to get unique path extension from
        :return: full path
        """
        if saver.path == "":
            return path_prefix
        return "{}.{}".format(path_prefix, saver.path)





#
# Copyright (c) 2017 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List, Union, Tuple

from rl_coach.base_parameters import AgentParameters, VisualizationParameters, TaskParameters, \
    PresetValidationParameters
from rl_coach.core_types import EnvironmentSteps
from rl_coach.environments.environment import EnvironmentParameters, Environment
from rl_coach.graph_managers.graph_manager import GraphManager, ScheduleParameters
from rl_coach.level_manager import LevelManager
from rl_coach.utils import short_dynamic_import


class HRLGraphManager(GraphManager):
    """
    A simple HRL graph manager creates a deep hierarchy with a single composite agent per hierarchy level, and a single
    environment which is interacted with.
    """
    def __init__(self, agents_params: List[AgentParameters], env_params: EnvironmentParameters,
                 schedule_params: ScheduleParameters, vis_params: VisualizationParameters,
                 consecutive_steps_to_run_each_level: Union[EnvironmentSteps, List[EnvironmentSteps]],
                 preset_validation_params: PresetValidationParameters = PresetValidationParameters()):
        """
        :param agents_params: the parameters of all the agents in the hierarchy starting from the top level of the
                              hierarchy to the bottom level
        :param env_params: the parameters of the environment
        :param schedule_params: the parameters for scheduling the graph
        :param vis_params: the visualization parameters
        :param consecutive_steps_to_run_each_level: the number of time steps that each level is ran.
            for example, when the top level gives the bottom level a goal, the bottom level can act for
            consecutive_steps_to_run_each_level steps and try to reach that goal. This is expected to be either
            an EnvironmentSteps which will be used for all levels, or an EnvironmentSteps for each level as a list.
        """
        super().__init__('hrl_graph', schedule_params, vis_params)
        self.agents_params = agents_params
        self.env_params = env_params
        self.preset_validation_params = preset_validation_params
        if isinstance(consecutive_steps_to_run_each_level, list):
            if len(consecutive_steps_to_run_each_level) != len(self.agents_params):
                raise ValueError("If the consecutive_steps_to_run_each_level is given as a list, it should match "
                                 "the number of levels in the hierarchy. Alternatively, it is possible to use a single "
                                 "value for all the levels, by passing an EnvironmentSteps")
        elif isinstance(consecutive_steps_to_run_each_level, EnvironmentSteps):
            self.consecutive_steps_to_run_each_level = [consecutive_steps_to_run_each_level] * len(self.agents_params)

        for agent_params in agents_params:
            agent_params.visualization = self.visualization_parameters
            if agent_params.input_filter is None:
                agent_params.input_filter = self.env_params.default_input_filter()
            if agent_params.output_filter is None:
                agent_params.output_filter = self.env_params.default_output_filter()

        if len(self.agents_params) < 2:
            raise ValueError("The HRL graph manager must receive the agent parameters for at least two levels of the "
                             "hierarchy. Otherwise, use the basic RL graph manager.")

    def _create_graph(self, task_parameters: TaskParameters) -> Tuple[List[LevelManager], List[Environment]]:
        self.env_params.seed = task_parameters.seed
        env = short_dynamic_import(self.env_params.path)(**self.env_params.__dict__,
                                                         visualization_parameters=self.visualization_parameters)

        for agent_params in self.agents_params:
            agent_params.task_parameters = task_parameters

        # we need to build the hierarchy in reverse order (from the bottom up) in order for the spaces of each level
        # to be known
        level_managers = []
        current_env = env
        # out_action_space = env.action_space
        for level_idx, agent_params in reversed(list(enumerate(self.agents_params))):
            # TODO: the code below is specific for HRL on observation scale
            # in action space
            # if level_idx == 0:
            #     # top level agents do not get directives
            #     in_action_space = None
            # else:
            #     pass

                # attention_size = (env.state_space['observation'].shape - 1)//4
                # in_action_space = AttentionActionSpace(shape=2, low=0, high=env.state_space['observation'].shape - 1,
                #                             forced_attention_size=attention_size)
                # agent_params.output_filter.action_filters['masking'].set_masking(0, attention_size)

            agent_params.name = "agent_{}".format(level_idx)
            agent_params.is_a_highest_level_agent = level_idx == 0
            agent = short_dynamic_import(agent_params.path)(agent_params)

            level_manager = LevelManager(
                agents=agent,
                environment=current_env,
                real_environment=env,
                steps_limit=self.consecutive_steps_to_run_each_level[level_idx],
                should_reset_agent_state_after_time_limit_passes=level_idx > 0,
                name="level_{}".format(level_idx)
            )
            current_env = level_manager
            level_managers.insert(0, level_manager)

            # out_action_space = in_action_space

        return level_managers, [env]





#
# Copyright (c) 2017 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Module providing helper classes and functions for reading/writing checkpoint state
"""

import os
import re
from typing import List, Union, Tuple


class SingleCheckpoint(object):
    """
    Helper class for storing checkpoint name and number
    """
    def __init__(self, num: int, name: str):
        """
        :param num: checkpoint number
        :param name: checkpoint name (i.e. the prefix for all checkpoint files)
        """
        self._num = num
        self._name = name

    @property
    def num(self) -> int:
        return self._num

    @property
    def name(self) -> str:
        return self._name

    def __str__(self):
        return self._name

    def __repr__(self):
        return str(self)

    def __eq__(self, other: 'SingleCheckpoint'):
        if not isinstance(other, SingleCheckpoint):
            return False
        return self._name == other._name and self._num == other._num

    def __ne__(self, other):
        return not self.__eq__(other)


class CheckpointState(object):
    """
    Helper class for checkpoint directory information. It replicates
    the CheckpointState protobuf class in tensorflow with addition of
    two new functions: last_checkpoint() and all_checkpoints()
    """
    def __init__(self, checkpoints: List[SingleCheckpoint], checkpoint_dir: str):
        """
        :param checkpoints: sorted list of checkpoints from oldest to newest. checkpoint[-1] is
            considered to be the most recent checkpoint.
        :param checkpoint_dir: checkpoint directory which is added to the paths
        """
        self._checkpoints = checkpoints
        self._checkpoin_dir = checkpoint_dir

    @property
    def all_checkpoints(self) -> List[SingleCheckpoint]:
        """
        :return: list of all checkpoints
        """
        return self._checkpoints

    @property
    def last_checkpoint(self) -> SingleCheckpoint:
        """
        :return: the most recent checkpoint
        """
        return self._checkpoints[-1]

    @property
    def all_model_checkpoint_paths(self) -> List[str]:
        """
        TF compatible function call to get all checkpoints
        :return: list of all available model checkpoint paths
        """
        return [os.path.join(self._checkpoin_dir, c.name) for c in self._checkpoints]

    @property
    def model_checkpoint_path(self) -> str:
        """
        TF compatible call to get most recent checkpoint
        :return: path of the most recent model checkpoint
        """
        return os.path.join(self._checkpoin_dir, self._checkpoints[-1].name)

    def __str__(self):
        out_str = 'model_checkpoint_path: {}\n'.format(self.model_checkpoint_path)
        for c in self.all_model_checkpoint_paths:
            out_str += 'all_model_checkpoint_paths: {}\n'.format(c)
        return out_str

    def __repr__(self):
        return str(self._checkpoints)


class CheckpointStateFile(object):
    """
    Helper class for reading from and writing to the checkpoint state file
    """
    checkpoint_state_filename = '.coach_checkpoint'

    def __init__(self, checkpoint_dir: str):
        self._checkpoint_state_path = os.path.join(checkpoint_dir, self.checkpoint_state_filename)

    def exists(self) -> bool:
        """
        :return: True if checkpoint state file exists, false otherwise
        """
        return os.path.exists(self._checkpoint_state_path)

    def read(self) -> Union[None, SingleCheckpoint]:
        """
        Read checkpoint state file and interpret its content
        :return:
        """
        if not self.exists():
            return None
        with open(self._checkpoint_state_path, 'r') as fd:
            return CheckpointFilenameParser().parse(fd.read(256))

    def write(self, data: SingleCheckpoint) -> None:
        """
        Writes data to checkpoint state file
        :param data: string data
        """
        with open(self._checkpoint_state_path, 'w') as fd:
            fd.write(data.name)

    @property
    def filename(self) -> str:
        return self.checkpoint_state_filename

    @property
    def path(self) -> str:
        return self._checkpoint_state_path


class CheckpointStateReader(object):
    """
    Class for scanning checkpoint directory and updating the checkpoint state
    """
    def __init__(self, checkpoint_dir: str, checkpoint_state_optional: bool=True):
        """
        :param checkpoint_dir: path to checkpoint directory
        :param checkpoint_state_optional: If True, checkpoint state file is optional and if not found,
            directory is scanned to find the latest checkpoint. Default is True for backward compatibility
        """
        self._checkpoint_dir = checkpoint_dir
        self._checkpoint_state_file = CheckpointStateFile(self._checkpoint_dir)
        self._checkpoint_state_optional = checkpoint_state_optional

    def get_latest(self) -> SingleCheckpoint:
        """
        Tries to read the checkpoint state file. If that fails, discovers latest by reading the entire directory.
        :return: checkpoint object representing the latest checkpoint
        """
        latest = self._checkpoint_state_file.read()
        if latest is None and self._checkpoint_state_optional:
            all_checkpoints = _filter_checkpoint_files(os.listdir(self._checkpoint_dir))
            if len(all_checkpoints) > 0:
                latest = all_checkpoints[-1]
        return latest

    def get_all(self) -> List[SingleCheckpoint]:
        """
        Reads both the checkpoint state file as well as contents of the directory and merges them into one list.
        :return: list of checkpoint objects
        """
        # discover all checkpoint files in directory if requested or if a valid checkpoint-state file doesn't exist
        all_checkpoints = _filter_checkpoint_files(os.listdir(self._checkpoint_dir))
        last_checkpoint = self._checkpoint_state_file.read()
        if last_checkpoint is not None:
            # remove excess checkpoints: higher checkpoint number, but not recent (e.g. from a previous run)
            all_checkpoints = all_checkpoints[: all_checkpoints.index(last_checkpoint) + 1]
        elif not self._checkpoint_state_optional:
            # if last_checkpoint is not discovered from the checkpoint-state file and it isn't optional, then
            # all checkpoint files discovered must be partial or invalid, so don't return anything
            all_checkpoints.clear()
        return all_checkpoints


class CheckpointStateUpdater(object):
    """
    Class for scanning checkpoint directory and updating the checkpoint state
    """
    def __init__(self, checkpoint_dir: str, read_all: bool=False):
        """
        :param checkpoint_dir: path to checkpoint directory
        :param read_all: whether to scan the directory for existing checkpoints
        """
        self._checkpoint_dir = checkpoint_dir
        self._checkpoint_state_file = CheckpointStateFile(checkpoint_dir)
        self._all_checkpoints = list()
        # Read checkpoint state and initialize
        state_reader = CheckpointStateReader(checkpoint_dir)
        if read_all:
            self._all_checkpoints = state_reader.get_all()
        else:
            latest = state_reader.get_latest()
            if latest is not None:
                self._all_checkpoints = [latest]

    def update(self, checkpoint: SingleCheckpoint) -> None:
        """
        Update the checkpoint state with the latest checkpoint.
        :param checkpoint: SingleCheckpoint object containing name and number of checkpoint
        """
        self._all_checkpoints.append(checkpoint)
        # Simply write checkpoint_name to checkpoint-state file
        self._checkpoint_state_file.write(checkpoint)

    @property
    def last_checkpoint(self) -> Union[None, SingleCheckpoint]:
        if len(self._all_checkpoints) == 0:
            return None
        return self._all_checkpoints[-1]

    @property
    def all_checkpoints(self) -> List[SingleCheckpoint]:
        return self._all_checkpoints

    def get_checkpoint_state(self) -> Union[None, CheckpointState]:
        """
        :return: The most recent checkpoint state
        """
        if len(self._all_checkpoints) == 0:
            return None
        return CheckpointState(self._all_checkpoints, self._checkpoint_dir)


class CheckpointFilenameParser(object):
    """
    Helper object for parsing filenames that are potentially checkpoints
    """
    coach_checkpoint_filename_pattern = r'\A(([0-9]+)[^0-9])?.*?\.ckpt(-([0-9]+))?'

    def __init__(self):
        self._prog = re.compile(self.coach_checkpoint_filename_pattern)

    def parse(self, filename: str) -> Union[None, SingleCheckpoint]:
        """
        Tries to parse the filename using the checkpoint filename pattern. If successful,
        it returns tuple of (checkpoint-number, checkpoint-name). Otherwise it returns None.
        :param filename: filename to be parsed
        :return: None or (checkpoint-number, checkpoint-name)
        """
        m = self._prog.search(filename)
        if m is not None and (m.group(2) is not None or m.group(4) is not None):
            assert m.group(2) is None or m.group(4) is None  # Only one group must be valid
            checkpoint_num = int(m.group(2) if m.group(2) is not None else m.group(4))
            return SingleCheckpoint(checkpoint_num, m.group(0))
        return None


def _filter_checkpoint_files(filenames: List[str], sort_by_num: bool=True) -> List[SingleCheckpoint]:
    """
    Given a list of potential file names, return the ones that match checkpoint pattern along with
    the checkpoint number of each file name.
    :param filenames: list of all filenames
    :param sort_by_num: whether to sort the output result by checkpoint number
    :return: list of (checkpoint-number, checkpoint-filename) tuples
    """
    parser = CheckpointFilenameParser()
    checkpoints = [ckp for ckp in [parser.parse(fn) for fn in filenames] if ckp is not None]
    if sort_by_num:
        checkpoints.sort(key=lambda x: x.num)
    return checkpoints


def get_checkpoint_state(checkpoint_dir: str, all_checkpoints=False) ->Union[CheckpointState, None]:
    """
    Scan checkpoint directory and find the list of checkpoint files.
    :param checkpoint_dir: directory where checkpoints are saved
    :param all_checkpoints: if True, scan the directory and return list of all checkpoints
        as well as the most recent one
    :return: a CheckpointState for checkpoint_dir containing a sorted list of checkpoints by checkpoint-number.
        If no matching files are found, returns None.
    """
    return CheckpointStateUpdater(checkpoint_dir, read_all=all_checkpoints).get_checkpoint_state()



#
# Copyright (c) 2017 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import pickle

import numpy as np
import tensorflow as tf

from rl_coach.utilities.shared_running_stats import SharedRunningStats


class TFSharedRunningStats(SharedRunningStats):
    def __init__(self, replicated_device=None, epsilon=1e-2, name="", create_ops=True, pubsub_params=None):
        super().__init__(name=name, pubsub_params=pubsub_params)
        self.sess = None
        self.replicated_device = replicated_device
        self.epsilon = epsilon
        self.ops_were_created = False
        if create_ops:
            with tf.device(replicated_device):
                self.set_params()

    def set_params(self, shape=[1], clip_values=None):
        """
        set params and create ops

        :param shape: shape of the stats to track
        :param clip_values: if not None, sets clip min/max thresholds
        """

        self.clip_values = clip_values
        with tf.variable_scope(self.name):
            self._sum = tf.get_variable(
                dtype=tf.float64,
                initializer=tf.constant_initializer(0.0),
                name="running_sum", trainable=False, shape=shape, validate_shape=False,
                collections=[tf.GraphKeys.GLOBAL_VARIABLES])
            self._sum_squares = tf.get_variable(
                dtype=tf.float64,
                initializer=tf.constant_initializer(self.epsilon),
                name="running_sum_squares", trainable=False, shape=shape, validate_shape=False,
                collections=[tf.GraphKeys.GLOBAL_VARIABLES])
            self._count = tf.get_variable(
                dtype=tf.float64,
                shape=(),
                initializer=tf.constant_initializer(self.epsilon),
                name="count", trainable=False, collections=[tf.GraphKeys.GLOBAL_VARIABLES])

            self._shape = None
            self._mean = tf.div(self._sum, self._count, name="mean")
            self._std = tf.sqrt(tf.maximum((self._sum_squares - self._count * tf.square(self._mean))
                                           / tf.maximum(self._count-1, 1), self.epsilon), name="stdev")
            self.tf_mean = tf.cast(self._mean, 'float32')
            self.tf_std = tf.cast(self._std, 'float32')

            self.new_sum = tf.placeholder(dtype=tf.float64, name='sum')
            self.new_sum_squares = tf.placeholder(dtype=tf.float64, name='var')
            self.newcount = tf.placeholder(shape=[], dtype=tf.float64, name='count')

            self._inc_sum = tf.assign_add(self._sum, self.new_sum, use_locking=True)
            self._inc_sum_squares = tf.assign_add(self._sum_squares, self.new_sum_squares, use_locking=True)
            self._inc_count = tf.assign_add(self._count, self.newcount, use_locking=True)

            self.raw_obs = tf.placeholder(dtype=tf.float64, name='raw_obs')
            self.normalized_obs = (self.raw_obs - self._mean) / self._std
            if self.clip_values is not None:
                self.clipped_obs = tf.clip_by_value(self.normalized_obs, self.clip_values[0], self.clip_values[1])

            self.ops_were_created = True

    def set_session(self, sess):
        self.sess = sess

    def push_val(self, x):
        x = x.astype('float64')
        self.sess.run([self._inc_sum, self._inc_sum_squares, self._inc_count],
                      feed_dict={
                          self.new_sum: x.sum(axis=0).ravel(),
                          self.new_sum_squares: np.square(x).sum(axis=0).ravel(),
                          self.newcount: np.array(len(x), dtype='float64')
                     })
        if self._shape is None:
            self._shape = x.shape

    @property
    def n(self):
        return self.sess.run(self._count)

    @property
    def mean(self):
        return self.sess.run(self._mean)

    @property
    def var(self):
        return self.std ** 2

    @property
    def std(self):
        return self.sess.run(self._std)

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, val):
        self._shape = val
        self.new_sum.set_shape(val)
        self.new_sum_squares.set_shape(val)
        self.tf_mean.set_shape(val)
        self.tf_std.set_shape(val)
        self._sum.set_shape(val)
        self._sum_squares.set_shape(val)

    def normalize(self, batch):
        if self.clip_values is not None:
            return self.sess.run(self.clipped_obs, feed_dict={self.raw_obs: batch})
        else:
            return self.sess.run(self.normalized_obs, feed_dict={self.raw_obs: batch})

    def save_state_to_checkpoint(self, checkpoint_dir: str, checkpoint_prefix: str):
        # Since the internal state is maintained as part of the TF graph, no need to do anything special for
        # save/restore, when going from single-node-multi-thread run back to a single-node-multi-worker run.
        # Nevertheless, if we'll want to restore a checkpoint back to either a * single-worker, or a
        # multi-node-multi-worker * run, we have to save the internal state, so that it can be restored to the
        # NumpySharedRunningStats implementation.

        dict_to_save = {'_mean': self.mean,
                        '_std': self.std,
                        '_count': self.n,
                        '_sum': self.sess.run(self._sum),
                        '_sum_squares': self.sess.run(self._sum_squares)}

        with open(os.path.join(checkpoint_dir, str(checkpoint_prefix) + '.srs'), 'wb') as f:
            pickle.dump(dict_to_save, f, pickle.HIGHEST_PROTOCOL)

    def restore_state_from_checkpoint(self, checkpoint_dir: str, checkpoint_prefix: str):
        # Since the internal state is maintained as part of the TF graph, no need to do anything special for
        # save/restore, when going from single-node-multi-thread run back to a single-node-multi-worker run.
        # Restoring from either a * single-worker, or a multi-node-multi-worker * run, to a single-node-multi-thread run
        # is not supported.
        pass



#
# Copyright (c) 2017 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import math
from types import FunctionType
import tensorflow as tf

from rl_coach.architectures import layers
from rl_coach.architectures.tensorflow_components import utils


def batchnorm_activation_dropout(input_layer, batchnorm, activation_function, dropout_rate, is_training, name):
    layers = [input_layer]

    # batchnorm
    if batchnorm:
        layers.append(
            tf.layers.batch_normalization(layers[-1], name="{}_batchnorm".format(name), training=is_training)
        )

    # activation
    if activation_function:
        if isinstance(activation_function, str):
            activation_function = utils.get_activation_function(activation_function)
        layers.append(
            activation_function(layers[-1], name="{}_activation".format(name))
        )

    # dropout
    if dropout_rate > 0:
        layers.append(
            tf.layers.dropout(layers[-1], dropout_rate, name="{}_dropout".format(name), training=is_training)
        )

    # remove the input layer from the layers list
    del layers[0]

    return layers


# define global dictionary for storing layer type to layer implementation mapping
tf_layer_dict = dict()
tf_layer_class_dict = dict()


def reg_to_tf_instance(layer_type) -> FunctionType:
    """ function decorator that registers layer implementation
    :return: decorated function
    """
    def reg_impl_decorator(func):
        assert layer_type not in tf_layer_dict
        tf_layer_dict[layer_type] = func
        return func
    return reg_impl_decorator


def reg_to_tf_class(layer_type) -> FunctionType:
    """ function decorator that registers layer type
    :return: decorated function
    """
    def reg_impl_decorator(func):
        assert layer_type not in tf_layer_class_dict
        tf_layer_class_dict[layer_type] = func
        return func
    return reg_impl_decorator


def convert_layer(layer):
    """
    If layer instance is callable (meaning this is already a concrete TF class), return layer, otherwise convert to TF type
    :param layer: layer to be converted
    :return: converted layer if not callable, otherwise layer itself
    """
    if callable(layer):
        return layer
    return tf_layer_dict[type(layer)](layer)


def convert_layer_class(layer_class):
    """
    If layer instance is callable, return layer, otherwise convert to TF type
    :param layer: layer to be converted
    :return: converted layer if not callable, otherwise layer itself
    """
    if hasattr(layer_class, 'to_tf_instance'):
        return layer_class
    else:
        return tf_layer_class_dict[layer_class]()


class Conv2d(layers.Conv2d):
    def __init__(self, num_filters: int, kernel_size: int, strides: int):
        super(Conv2d, self).__init__(num_filters=num_filters, kernel_size=kernel_size, strides=strides)

    def __call__(self, input_layer, name: str=None, is_training=None):
        """
        returns a tensorflow conv2d layer
        :param input_layer: previous layer
        :param name: layer name
        :return: conv2d layer
        """
        return tf.layers.conv2d(input_layer, filters=self.num_filters, kernel_size=self.kernel_size,
                                strides=self.strides, data_format='channels_last', name=name)

    @staticmethod
    @reg_to_tf_instance(layers.Conv2d)
    def to_tf_instance(base: layers.Conv2d):
            return Conv2d(
                num_filters=base.num_filters,
                kernel_size=base.kernel_size,
                strides=base.strides)

    @staticmethod
    @reg_to_tf_class(layers.Conv2d)
    def to_tf_class():
        return Conv2d


class BatchnormActivationDropout(layers.BatchnormActivationDropout):
    def __init__(self, batchnorm: bool=False, activation_function=None, dropout_rate: float=0):
        super(BatchnormActivationDropout, self).__init__(
            batchnorm=batchnorm, activation_function=activation_function, dropout_rate=dropout_rate)

    def __call__(self, input_layer, name: str=None, is_training=None):
        """
        returns a list of tensorflow batchnorm, activation and dropout layers
        :param input_layer: previous layer
        :param name: layer name
        :return: batchnorm, activation and dropout layers
        """
        return batchnorm_activation_dropout(input_layer, batchnorm=self.batchnorm,
                                            activation_function=self.activation_function,
                                            dropout_rate=self.dropout_rate,
                                            is_training=is_training, name=name)

    @staticmethod
    @reg_to_tf_instance(layers.BatchnormActivationDropout)
    def to_tf_instance(base: layers.BatchnormActivationDropout):
        return BatchnormActivationDropout, BatchnormActivationDropout(
                batchnorm=base.batchnorm,
                activation_function=base.activation_function,
                dropout_rate=base.dropout_rate)

    @staticmethod
    @reg_to_tf_class(layers.BatchnormActivationDropout)
    def to_tf_class():
        return BatchnormActivationDropout


class Dense(layers.Dense):
    def __init__(self, units: int):
        super(Dense, self).__init__(units=units)

    def __call__(self, input_layer, name: str=None, kernel_initializer=None, activation=None, is_training=None):
        """
        returns a tensorflow dense layer
        :param input_layer: previous layer
        :param name: layer name
        :return: dense layer
        """
        return tf.layers.dense(input_layer, self.units, name=name, kernel_initializer=kernel_initializer,
                               activation=activation)

    @staticmethod
    @reg_to_tf_instance(layers.Dense)
    def to_tf_instance(base: layers.Dense):
        return Dense(units=base.units)

    @staticmethod
    @reg_to_tf_class(layers.Dense)
    def to_tf_class():
        return Dense


class NoisyNetDense(layers.NoisyNetDense):
    """
    A factorized Noisy Net layer

    https://arxiv.org/abs/1706.10295.
    """

    def __init__(self, units: int):
        super(NoisyNetDense, self).__init__(units=units)

    def __call__(self, input_layer, name: str, kernel_initializer=None, activation=None, is_training=None):
        """
        returns a NoisyNet dense layer
        :param input_layer: previous layer
        :param name: layer name
        :param kernel_initializer: initializer for kernels. Default is to use Gaussian noise that preserves stddev.
        :param activation: the activation function
        :return: dense layer
        """
        #TODO: noise sampling should be externally controlled. DQN is fine with sampling noise for every
        #      forward (either act or train, both for online and target networks).
        #      A3C, on the other hand, should sample noise only when policy changes (i.e. after every t_max steps)

        def _f(values):
            return tf.sqrt(tf.abs(values)) * tf.sign(values)

        def _factorized_noise(inputs, outputs):
            # TODO: use factorized noise only for compute intensive algos (e.g. DQN).
            #      lighter algos (e.g. DQN) should not use it
            noise1 = _f(tf.random_normal((inputs, 1)))
            noise2 = _f(tf.random_normal((1, outputs)))
            return tf.matmul(noise1, noise2)

        num_inputs = input_layer.get_shape()[-1].value
        num_outputs = self.units

        stddev = 1 / math.sqrt(num_inputs)
        activation = activation if activation is not None else (lambda x: x)

        if kernel_initializer is None:
            kernel_mean_initializer = tf.random_uniform_initializer(-stddev, stddev)
            kernel_stddev_initializer = tf.random_uniform_initializer(-stddev * self.sigma0, stddev * self.sigma0)
        else:
            kernel_mean_initializer = kernel_stddev_initializer = kernel_initializer
        with tf.variable_scope(None, default_name=name):
            weight_mean = tf.get_variable('weight_mean', shape=(num_inputs, num_outputs),
                                          initializer=kernel_mean_initializer)
            bias_mean = tf.get_variable('bias_mean', shape=(num_outputs,), initializer=tf.zeros_initializer())

            weight_stddev = tf.get_variable('weight_stddev', shape=(num_inputs, num_outputs),
                                            initializer=kernel_stddev_initializer)
            bias_stddev = tf.get_variable('bias_stddev', shape=(num_outputs,),
                                          initializer=kernel_stddev_initializer)
            bias_noise = _f(tf.random_normal((num_outputs,)))
            weight_noise = _factorized_noise(num_inputs, num_outputs)

        bias = bias_mean + bias_stddev * bias_noise
        weight = weight_mean + weight_stddev * weight_noise
        return activation(tf.matmul(input_layer, weight) + bias)

    @staticmethod
    @reg_to_tf_instance(layers.NoisyNetDense)
    def to_tf_instance(base: layers.NoisyNetDense):
        return NoisyNetDense(units=base.units)

    @staticmethod
    @reg_to_tf_class(layers.NoisyNetDense)
    def to_tf_class():
        return NoisyNetDense



#
# Copyright (c) 2017 Intel Corporation 
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List

import numpy as np

from rl_coach.core_types import ActionType
from rl_coach.exploration_policies.exploration_policy import ExplorationPolicy, ExplorationParameters
from rl_coach.spaces import ActionSpace, DiscreteActionSpace, BoxActionSpace


class GreedyParameters(ExplorationParameters):
    @property
    def path(self):
        return 'rl_coach.exploration_policies.greedy:Greedy'


class Greedy(ExplorationPolicy):
    """
    The Greedy exploration policy is intended for both discrete and continuous action spaces.
    For discrete action spaces, it always selects the action with the maximum value, as given by the agent.
    For continuous action spaces, it always return the exact action, as it was given by the agent.
    """
    def __init__(self, action_space: ActionSpace):
        """
        :param action_space: the action space used by the environment
        """
        super().__init__(action_space)

    def get_action(self, action_values: List[ActionType]) -> ActionType:
        if type(self.action_space) == DiscreteActionSpace:
            return np.argmax(action_values)
        if type(self.action_space) == BoxActionSpace:
            return action_values

    def get_control_param(self):
        return 0



from typing import Union
from types import ModuleType

import mxnet as mx
from rl_coach.architectures.embedder_parameters import InputEmbedderParameters
from rl_coach.architectures.mxnet_components.embedders.embedder import InputEmbedder
from rl_coach.architectures.mxnet_components.layers import Conv2d
from rl_coach.base_parameters import EmbedderScheme

nd_sym_type = Union[mx.nd.NDArray, mx.sym.Symbol]


class ImageEmbedder(InputEmbedder):
    def __init__(self, params: InputEmbedderParameters):
        """
        An image embedder is an input embedder that takes an image input from the state and produces a vector
        embedding by passing it through a neural network.

        :param params: parameters object containing input_clipping, input_rescaling, batchnorm, activation_function
            and dropout properties.
        """
        super(ImageEmbedder, self).__init__(params)
        self.input_rescaling = params.input_rescaling['image']
        self.input_offset = params.input_offset['image']

    @property
    def schemes(self) -> dict:
        """
        Schemes are the pre-defined network architectures of various depths and complexities that can be used. Are used
        to create Block when ImageEmbedder is initialised.

        :return: dictionary of schemes, with key of type EmbedderScheme enum and value being list of mxnet.gluon.Block.
        """
        return {
            EmbedderScheme.Empty:
                [],

            EmbedderScheme.Shallow:
                [
                    Conv2d(num_filters=32, kernel_size=8, strides=4)
                ],

            # Use for Atari DQN
            EmbedderScheme.Medium:
                [
                    Conv2d(num_filters=32, kernel_size=8, strides=4),
                    Conv2d(num_filters=64, kernel_size=4, strides=2),
                    Conv2d(num_filters=64, kernel_size=3, strides=1)
                ],

            # Use for Carla
            EmbedderScheme.Deep:
                [
                    Conv2d(num_filters=32, kernel_size=5, strides=2),
                    Conv2d(num_filters=32, kernel_size=3, strides=1),
                    Conv2d(num_filters=64, kernel_size=3, strides=2),
                    Conv2d(num_filters=64, kernel_size=3, strides=1),
                    Conv2d(num_filters=128, kernel_size=3, strides=2),
                    Conv2d(num_filters=128, kernel_size=3, strides=1),
                    Conv2d(num_filters=256, kernel_size=3, strides=2),
                    Conv2d(num_filters=256, kernel_size=3, strides=1)
                ]
        }

    def hybrid_forward(self, F: ModuleType, x: nd_sym_type, *args, **kwargs) -> nd_sym_type:
        """
        Used for forward pass through embedder network.

        :param F: backend api, either `mxnet.nd` or `mxnet.sym` (if block has been hybridized).
        :param x: image representing environment state, of shape (batch_size, in_channels, height, width).
        :return: embedding of environment state, of shape (batch_size, channels).
        """
        # convert from NHWC to NCHW (default for MXNet Convolutions)
        x = x.transpose((0,3,1,2))
        return super(ImageEmbedder, self).hybrid_forward(F, x, *args, **kwargs)



#
# Copyright (c) 2017 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from collections import OrderedDict
from typing import Union

from rl_coach.agents.agent import Agent
from rl_coach.core_types import RunPhase, ActionInfo
from rl_coach.logger import screen
from rl_coach.spaces import DiscreteActionSpace


## This is an abstract agent - there is no learn_from_batch method ##

# Imitation Agent
class ImitationAgent(Agent):
    def __init__(self, agent_parameters, parent: Union['LevelManager', 'CompositeAgent']=None):
        super().__init__(agent_parameters, parent)
        self.imitation = True

    def extract_action_values(self, prediction):
        return prediction.squeeze()

    def choose_action(self, curr_state):
        # convert to batch so we can run it through the network
        prediction = self.networks['main'].online_network.predict(self.prepare_batch_for_inference(curr_state, 'main'))

        # get action values and extract the best action from it
        action_values = self.extract_action_values(prediction)
        self.exploration_policy.change_phase(RunPhase.TEST)
        action = self.exploration_policy.get_action(action_values)
        action_info = ActionInfo(action=action)

        return action_info

    def log_to_screen(self):
        # log to screen
        if self.phase == RunPhase.TRAIN:
            # for the training phase - we log during the episode to visualize the progress in training
            log = OrderedDict()
            if self.task_id is not None:
                log["Worker"] = self.task_id
            log["Episode"] = self.current_episode
            log["Loss"] = self.loss.values[-1]
            log["Training iteration"] = self.training_iteration
            screen.log_dict(log, prefix="Training")
        else:
            # for the evaluation phase - logging as in regular RL
            super().log_to_screen()

    def learn_from_batch(self, batch):
        raise NotImplementedError("ImitationAgent is an abstract agent. Not to be used directly.")




from configparser import ConfigParser
import pytest
import argparse
import os
from rl_coach.coach import CoachLauncher
import sys
from minio import Minio


def generate_config(image, memory_backend, s3_end_point, s3_bucket_name, s3_creds_file, config_file):
    """
    Generate the s3 config file to be used and also the dist-coach-config.template to be used for the test
    It reads the `AWS_ACCESS_KEY_ID` and `AWS_SECRET_ACCESS_KEY` env vars and fails if they are not provided.
    """
    # Write s3 creds
    aws_config = ConfigParser({
        'aws_access_key_id': os.environ.get('AWS_ACCESS_KEY_ID'),
        'aws_secret_access_key': os.environ.get('AWS_SECRET_ACCESS_KEY')
    }, default_section='default')
    with open(s3_creds_file, 'w') as f:
        aws_config.write(f)

    coach_config = ConfigParser({
        'image': image,
        'memory_backend': memory_backend,
        'data_store': 's3',
        's3_end_point': s3_end_point,
        's3_bucket_name': s3_bucket_name,
        's3_creds_file': s3_creds_file
    }, default_section="coach")
    with open(config_file, 'w') as f:
        coach_config.write(f)


def test_command(command):
    """
    Launches the actual training.
    """
    sys.argv = command
    launcher = CoachLauncher()
    with pytest.raises(SystemExit) as e:
        launcher.launch()
    assert e.value.code == 0


def clear_bucket(s3_end_point, s3_bucket_name):
    """
    Clear the bucket before starting the test.
    """
    access_key = os.environ.get('AWS_ACCESS_KEY_ID')
    secret_access_key = os.environ.get('AWS_SECRET_ACCESS_KEY')
    minio_client = Minio(s3_end_point, access_key=access_key, secret_key=secret_access_key)
    try:
        for obj in minio_client.list_objects_v2(s3_bucket_name, recursive=True):
            minio_client.remove_object(s3_bucket_name, obj.object_name)
    except Exception:
        pass


def test_dc(command, image, memory_backend, s3_end_point, s3_bucket_name, s3_creds_file, config_file):
    """
    Entry point into the test
    """
    clear_bucket(s3_end_point, s3_bucket_name)
    command = command.format(template=config_file).split(' ')
    test_command(command)


def get_tests():
    """
    All the presets to test. New presets should be added here.
    """
    tests = [
        'rl_coach/coach.py -p CartPole_ClippedPPO -dc -e sample -dcp {template} --dump_worker_logs -asc --is_multi_node_test --seed 1',
        'rl_coach/coach.py -p Mujoco_ClippedPPO -lvl inverted_pendulum -dc -e sample -dcp {template} --dump_worker_logs -asc --is_multi_node_test --seed 1'
    ]
    return tests


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '-i', '--image', help="(string) Name of the testing image", type=str, default=None
    )
    parser.add_argument(
        '-mb', '--memory_backend', help="(string) Name of the memory backend", type=str, default="redispubsub"
    )
    parser.add_argument(
        '-e', '--endpoint', help="(string) Name of the s3 endpoint", type=str, default='s3.amazonaws.com'
    )
    parser.add_argument(
        '-cr', '--creds_file', help="(string) Path of the s3 creds file", type=str, default='.aws_creds'
    )
    parser.add_argument(
        '-b', '--bucket', help="(string) Name of the bucket for s3", type=str, default=None
    )

    args = parser.parse_args()

    if not args.bucket:
        print("bucket_name required for s3")
        exit(1)
    if not os.environ.get('AWS_ACCESS_KEY_ID') or not os.environ.get('AWS_SECRET_ACCESS_KEY'):
        print("AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY env vars need to be set")
        exit(1)

    config_file = './tmp.cred'
    generate_config(args.image, args.memory_backend, args.endpoint, args.bucket, args.creds_file, config_file)
    for command in get_tests():
        test_dc(command, args.image, args.memory_backend, args.endpoint, args.bucket, args.creds_file, config_file)


if __name__ == "__main__":
    main()



import mxnet as mx
import os
import pytest
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))


from rl_coach.base_parameters import MiddlewareScheme
from rl_coach.architectures.middleware_parameters import FCMiddlewareParameters
from rl_coach.architectures.mxnet_components.middlewares.fc_middleware import FCMiddleware


@pytest.mark.unit_test
def test_fc_middleware():
    params = FCMiddlewareParameters(scheme=MiddlewareScheme.Medium)
    mid = FCMiddleware(params=params)
    mid.initialize()
    embedded_data = mx.nd.random.uniform(low=0, high=1, shape=(10, 100))
    output = mid(embedded_data)
    assert output.ndim == 2  # since last block was flatten
    assert output.shape[0] == 10  # since batch_size is 10
    assert output.shape[1] == 512  # since last layer of middleware (middle scheme) had 512 units



import os
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))

import pytest

from rl_coach.filters.observation.observation_rescale_to_size_filter import ObservationRescaleToSizeFilter, RescaleInterpolationType
from rl_coach.filters.observation.observation_crop_filter import ObservationCropFilter
from rl_coach.filters.reward.reward_clipping_filter import RewardClippingFilter
from rl_coach.filters.observation.observation_stacking_filter import ObservationStackingFilter
from rl_coach.filters.filter import InputFilter
from rl_coach.spaces import ImageObservationSpace
import numpy as np
from rl_coach.core_types import EnvResponse
from collections import OrderedDict


@pytest.mark.filterwarnings('ignore:Conversion of')
@pytest.mark.unit_test
def test_filter_stacking():
    # test that filter stacking works fine by taking as input a transition with:
    # - an observation of shape 210x160,
    # - a reward of 100
    # filtering it by:
    # - rescaling the observation to 110x84
    # - cropping the observation to 84x84
    # - clipping the reward to 1
    # - stacking 4 observations to get 84x84x4

    env_response = EnvResponse({'observation': np.ones([210, 160])}, reward=100, game_over=False)

    filter1 = ObservationRescaleToSizeFilter(
        output_observation_space=ImageObservationSpace(np.array([110, 84]), high=255),
        rescaling_interpolation_type=RescaleInterpolationType.BILINEAR
    )

    filter2 = ObservationCropFilter(
        crop_low=np.array([16, 0]),
        crop_high=np.array([100, 84])
    )

    filter3 = RewardClippingFilter(
        clipping_low=-1,
        clipping_high=1
    )

    output_filter = ObservationStackingFilter(
        stack_size=4,
        stacking_axis=-1
    )

    input_filter = InputFilter(
        observation_filters={
            "observation": OrderedDict([
                ("filter1", filter1),
                ("filter2", filter2),
                ("output_filter", output_filter)
            ])},
        reward_filters=OrderedDict([
            ("filter3", filter3)
        ])
    )

    result = input_filter.filter(env_response)[0]
    observation = np.array(result.next_state['observation'])
    assert observation.shape == (84, 84, 4)
    assert np.all(observation == np.ones([84, 84, 4]))
    assert result.reward == 1





# nasty hack to deal with issue #46
import os
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))

import pytest
import numpy as np

from rl_coach.core_types import Transition
from rl_coach.memories.episodic.single_episode_buffer import SingleEpisodeBuffer


@pytest.fixture()
def buffer():
    return SingleEpisodeBuffer()


@pytest.mark.unit_test
def test_store_and_get(buffer: SingleEpisodeBuffer):
    # store single non terminal transition
    transition = Transition(state={"observation": np.array([1, 2, 3])}, action=1, reward=1, game_over=False)
    buffer.store(transition)
    assert buffer.length() == 1
    assert buffer.num_complete_episodes() == 0
    assert buffer.num_transitions_in_complete_episodes() == 0
    assert buffer.num_transitions() == 1

    # get the single stored transition
    episode = buffer.get(0)
    assert episode.length() == 1
    assert episode.get_first_transition() is transition    # check addresses are the same
    assert episode.get_last_transition() is transition   # check addresses are the same

    # store single terminal transition
    transition = Transition(state={"observation": np.array([1, 2, 3])}, action=1, reward=1, game_over=True)
    buffer.store(transition)
    assert buffer.length() == 1
    assert buffer.num_complete_episodes() == 1
    assert buffer.num_transitions_in_complete_episodes() == 2

    # check that the episode is valid
    episode = buffer.get(0)
    assert episode.length() == 2
    assert episode.get_transition(0).n_step_discounted_rewards == 1 + 0.99
    assert episode.get_transition(1).n_step_discounted_rewards == 1
    assert buffer.mean_reward() == 1

    # only one episode in the replay buffer
    episode = buffer.get(1)
    assert episode is None

    # adding transitions after the first episode was closed
    transition = Transition(state={"observation": np.array([1, 2, 3])}, action=1, reward=0, game_over=False)
    buffer.store(transition)
    assert buffer.length() == 1
    assert buffer.num_complete_episodes() == 0
    assert buffer.num_transitions_in_complete_episodes() == 0

    # still only one episode
    assert buffer.get(1) is None
    assert buffer.mean_reward() == 0


@pytest.mark.unit_test
def test_clean(buffer: SingleEpisodeBuffer):
    # add several transitions and then clean the buffer
    transition = Transition(state={"observation": np.array([1, 2, 3])}, action=1, reward=1, game_over=False)
    for i in range(10):
        buffer.store(transition)
    assert buffer.num_transitions() == 10
    buffer.clean()
    assert buffer.num_transitions() == 0

    # add more transitions after the clean and make sure they were really cleaned
    transition = Transition(state={"observation": np.array([1, 2, 3])}, action=1, reward=1, game_over=True)
    buffer.store(transition)
    assert buffer.num_transitions() == 1
    assert buffer.num_transitions_in_complete_episodes() == 1
    assert buffer.num_complete_episodes() == 1
    for i in range(10):
        assert buffer.sample(1)[0] is transition



#
# Copyright (c) 2017 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import tensorflow as tf

from rl_coach.architectures.tensorflow_components.layers import batchnorm_activation_dropout, Dense
from rl_coach.architectures.tensorflow_components.heads.head import Head
from rl_coach.base_parameters import AgentParameters
from rl_coach.core_types import ActionProbabilities
from rl_coach.spaces import SpacesDefinition


class DDPGActor(Head):
    def __init__(self, agent_parameters: AgentParameters, spaces: SpacesDefinition, network_name: str,
                 head_idx: int = 0, loss_weight: float = 1., is_local: bool = True, activation_function: str='tanh',
                 batchnorm: bool=True, dense_layer=Dense):
        super().__init__(agent_parameters, spaces, network_name, head_idx, loss_weight, is_local, activation_function,
                         dense_layer=dense_layer)
        self.name = 'ddpg_actor_head'
        self.return_type = ActionProbabilities

        self.num_actions = self.spaces.action.shape

        self.batchnorm = batchnorm

        # bounded actions
        self.output_scale = self.spaces.action.max_abs_range

        # a scalar weight that penalizes high activation values (before the activation function) for the final layer
        if hasattr(agent_parameters.algorithm, 'action_penalty'):
            self.action_penalty = agent_parameters.algorithm.action_penalty

    def _build_module(self, input_layer):
        # mean
        pre_activation_policy_values_mean = self.dense_layer(self.num_actions)(input_layer, name='fc_mean')
        policy_values_mean = batchnorm_activation_dropout(input_layer=pre_activation_policy_values_mean,
                                                          batchnorm=self.batchnorm,
                                                          activation_function=self.activation_function,
                                                          dropout_rate=0,
                                                          is_training=False,
                                                          name="BatchnormActivationDropout_0")[-1]
        self.policy_mean = tf.multiply(policy_values_mean, self.output_scale, name='output_mean')

        if self.is_local:
            # add a squared penalty on the squared pre-activation features of the action
            if self.action_penalty and self.action_penalty != 0:
                self.regularizations += \
                    [self.action_penalty * tf.reduce_mean(tf.square(pre_activation_policy_values_mean))]

        self.output = [self.policy_mean]

    def __str__(self):
        result = [
            'Dense (num outputs = {})'.format(self.num_actions[0])
        ]
        return '\n'.join(result)



#
# Copyright (c) 2017 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import tensorflow as tf

from rl_coach.architectures.tensorflow_components.layers import Dense
from rl_coach.architectures.tensorflow_components.heads.head import Head
from rl_coach.base_parameters import AgentParameters
from rl_coach.core_types import QActionStateValue
from rl_coach.spaces import SpacesDefinition, BoxActionSpace, DiscreteActionSpace
from rl_coach.utils import force_list


class RegressionHead(Head):
    def __init__(self, agent_parameters: AgentParameters, spaces: SpacesDefinition, network_name: str,
                 head_idx: int = 0, loss_weight: float = 1., is_local: bool = True, activation_function: str='relu',
                 dense_layer=Dense, scheme=[Dense(256), Dense(256)]):
        super().__init__(agent_parameters, spaces, network_name, head_idx, loss_weight, is_local, activation_function,
                         dense_layer=dense_layer)
        self.name = 'regression_head'
        self.scheme = scheme
        self.layers = []
        if isinstance(self.spaces.action, BoxActionSpace):
            self.num_actions = self.spaces.action.shape[0]
        elif isinstance(self.spaces.action, DiscreteActionSpace):
            self.num_actions = len(self.spaces.action.actions)
        self.return_type = QActionStateValue
        if agent_parameters.network_wrappers[self.network_name].replace_mse_with_huber_loss:
            self.loss_type = tf.losses.huber_loss
        else:
            self.loss_type = tf.losses.mean_squared_error

    def _build_module(self, input_layer):
        self.layers.append(input_layer)
        for idx, layer_params in enumerate(self.scheme):
            self.layers.extend(force_list(
                layer_params(input_layer=self.layers[-1], name='{}_{}'.format(layer_params.__class__.__name__, idx))
            ))

        self.layers.append(self.dense_layer(self.num_actions)(self.layers[-1], name='output'))
        self.output = self.layers[-1]

    def __str__(self):
        result = []
        for layer in self.layers:
            result.append(str(layer))
        return '\n'.join(result)




#
# Copyright (c) 2017 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import OrderedDict

import os
from genericpath import isdir, isfile
from os import listdir
from os.path import join
from enum import Enum
from bokeh.models import Div
from bokeh.plotting import curdoc
import tkinter as tk
from tkinter import filedialog
import colorsys

from rl_coach.core_types import TimeTypes

patches = {}
signals_files = {}
selected_file = None
x_axis = ['Episode #']
x_axis_options = [time_type.value.name for time_type in TimeTypes]
x_axis_labels = [time_type.value.label for time_type in TimeTypes]
current_color = 0

# spinner
root_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
with open(os.path.join(root_dir, 'dashboard_components/spinner.css'), 'r') as f:
    spinner_style = """<style>{}</style>""".format(f.read())
    spinner_html = """<ul class="spinner"><li></li><li></li><li></li><li></li>
                      <li>
                        <br>
                        <span style="font-size: 24px; font-weight: bold; margin-left: -175px; width: 400px; 
                        position: absolute; text-align: center;">
                            {}
                        </span>
                      </li></ul>"""
spinner = Div(text="""""")
displayed_doc = "landing_page"
layouts = {}


def generate_color_range(N, I):
    HSV_tuples = [(x*1.0/N, 0.5, I) for x in range(N)]
    RGB_tuples = map(lambda x: colorsys.hsv_to_rgb(*x), HSV_tuples)
    for_conversion = []
    for RGB_tuple in RGB_tuples:
        for_conversion.append((int(RGB_tuple[0]*255), int(RGB_tuple[1]*255), int(RGB_tuple[2]*255)))
    hex_colors = [rgb_to_hex(RGB_tuple) for RGB_tuple in for_conversion]
    return hex_colors, for_conversion


# convert RGB tuple to hexadecimal code
def rgb_to_hex(rgb):
    return '#%02x%02x%02x' % rgb


# convert hexadecimal to RGB tuple
def hex_to_dec(hex):
    red = ''.join(hex.strip('#')[0:2])
    green = ''.join(hex.strip('#')[2:4])
    blue = ''.join(hex.strip('#')[4:6])
    return int(red, 16), int(green, 16), int(blue,16)


color_resolution = 1000
brightness = 0.75  # change to have brighter/darker colors
crx = list(range(1, color_resolution+1))  # the resolution is 1000 colors
cry = [5 for i in range(len(crx))]
crcolor, crRGBs = generate_color_range(color_resolution, brightness)  # produce spectrum


def display_boards():
    global displayed_doc
    if displayed_doc == "landing_page":
        doc.remove_root(doc.roots[0])
        doc.add_root(layouts["boards"])
        displayed_doc = "boards"


def show_spinner(text="Loading..."):
    spinner.text = spinner_style + spinner_html.format(text)


def hide_spinner():
    spinner.text = ""


# takes path to dir and recursively adds all its files to paths
def add_directory_csv_files(dir_path, paths=None):
    if not paths:
        paths = []

    for p in listdir(dir_path):
        path = join(dir_path, p)
        if isdir(path):
            # call recursively for each dir
            paths = add_directory_csv_files(path, paths)
        elif isfile(path) and path.endswith('.csv'):
            # add every file to the list
            paths.append(path)

    return paths




class DialogApp():

    def getFileDialog(self):
        application_window = tk.Tk()

        # Build a list of tuples for each file type the file dialog should display
        my_filetypes = [('csv files', '.csv')]

        # Ask the user to select a one or more file names.
        answer = filedialog.askopenfilename(parent=application_window,
                                             initialdir=os.getcwd(),
                                             title="Please select a file",
                                             filetypes=my_filetypes)
        application_window.destroy()
        return answer


    def getDirDialog(self):
        application_window = tk.Tk()

        # Ask the user to select a folder.
        answer = filedialog.askdirectory(parent=application_window,
                                         initialdir=os.getcwd(),
                                         title="Please select a folder")
        application_window.destroy()
        return answer





class RunType(Enum):
    SINGLE_FOLDER_SINGLE_FILE = 1
    SINGLE_FOLDER_MULTIPLE_FILES = 2
    MULTIPLE_FOLDERS_SINGLE_FILES = 3
    MULTIPLE_FOLDERS_MULTIPLE_FILES = 4
    UNKNOWN = 0


class FolderType(Enum):
    SINGLE_FILE = 1
    MULTIPLE_FILES = 2
    MULTIPLE_FOLDERS = 3
    EMPTY = 4


dialog = DialogApp()


doc = curdoc()



from .q_head import QHead
from .categorical_q_head import CategoricalQHead
from .ddpg_actor_head import DDPGActor
from .dnd_q_head import DNDQHead
from .dueling_q_head import DuelingQHead
from .measurements_prediction_head import MeasurementsPredictionHead
from .naf_head import NAFHead
from .policy_head import PolicyHead
from .ppo_head import PPOHead
from .ppo_v_head import PPOVHead
from .quantile_regression_q_head import QuantileRegressionQHead
from .rainbow_q_head import RainbowQHead
from .v_head import VHead
from .acer_policy_head import ACERPolicyHead

__all__ = [
    'CategoricalQHead',
    'DDPGActor',
    'DNDQHead',
    'DuelingQHead',
    'MeasurementsPredictionHead',
    'NAFHead',
    'PolicyHead',
    'PPOHead',
    'PPOVHead',
    'QHead',
    'QuantileRegressionQHead',
    'RainbowQHead',
    'VHead',
    'ACERPolicyHead'
]



import os
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))

import pytest

from rl_coach.spaces import DiscreteActionSpace
from rl_coach.exploration_policies.e_greedy import EGreedy
from rl_coach.schedules import LinearSchedule
import numpy as np
from rl_coach.core_types import RunPhase


@pytest.mark.unit_test
def test_get_action():
    # discrete control
    action_space = DiscreteActionSpace(3)
    epsilon_schedule = LinearSchedule(1.0, 1.0, 1000)
    policy = EGreedy(action_space, epsilon_schedule, evaluation_epsilon=0)

    # verify that test phase gives greedy actions (evaluation_epsilon = 0)
    policy.change_phase(RunPhase.TEST)
    for i in range(100):
        best_action = policy.get_action(np.array([10, 20, 30]))
        assert best_action == 2

    # verify that train phase gives uniform actions (exploration = 1)
    policy.change_phase(RunPhase.TRAIN)
    counters = np.array([0, 0, 0])
    for i in range(30000):
        best_action = policy.get_action(np.array([10, 20, 30]))
        counters[best_action] += 1
    assert np.all(counters > 9500)  # this is noisy so we allow 5% error

    # TODO: test continuous actions


@pytest.mark.unit_test
def test_change_phase():
    # discrete control
    action_space = DiscreteActionSpace(3)
    epsilon_schedule = LinearSchedule(1.0, 0.1, 1000)
    policy = EGreedy(action_space, epsilon_schedule, evaluation_epsilon=0.01)

    # verify schedule not applying if not in training phase
    assert policy.get_control_param() == 1.0
    policy.change_phase(RunPhase.TEST)
    best_action = policy.get_action(np.array([10, 20, 30]))
    assert policy.epsilon_schedule.current_value == 1.0
    policy.change_phase(RunPhase.HEATUP)
    best_action = policy.get_action(np.array([10, 20, 30]))
    assert policy.epsilon_schedule.current_value == 1.0
    policy.change_phase(RunPhase.UNDEFINED)
    best_action = policy.get_action(np.array([10, 20, 30]))
    assert policy.epsilon_schedule.current_value == 1.0


@pytest.mark.unit_test
def test_get_control_param():
    # discrete control
    action_space = DiscreteActionSpace(3)
    epsilon_schedule = LinearSchedule(1.0, 0.1, 1000)
    policy = EGreedy(action_space, epsilon_schedule, evaluation_epsilon=0.01)

    # verify schedule applies to TRAIN phase
    policy.change_phase(RunPhase.TRAIN)
    for i in range(999):
        best_action = policy.get_action(np.array([10, 20, 30]))
        assert 1.0 > policy.get_control_param() > 0.1
    best_action = policy.get_action(np.array([10, 20, 30]))
    assert policy.get_control_param() == 0.1

    # test phases
    policy.change_phase(RunPhase.TEST)
    assert policy.get_control_param() == 0.01

    policy.change_phase(RunPhase.TRAIN)
    assert policy.get_control_param() == 0.1

    policy.change_phase(RunPhase.HEATUP)
    assert policy.get_control_param() == 0.1



#
# Copyright (c) 2019 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np


class DoublyRobust(object):

    @staticmethod
    def evaluate(ope_shared_stats: 'OpeSharedStats') -> tuple:
        """
        Run the off-policy evaluator to get a score for the goodness of the new policy, based on the dataset,
        which was collected using other policy(ies).

        Papers:
        https://arxiv.org/abs/1103.4601
        https://arxiv.org/pdf/1612.01205 (some more clearer explanations)

        :return: the evaluation score
        """

        ips = np.mean(ope_shared_stats.rho_all_dataset * ope_shared_stats.all_rewards)
        dm = np.mean(ope_shared_stats.all_v_values_reward_model_based)
        dr = np.mean(ope_shared_stats.rho_all_dataset *
                     (ope_shared_stats.all_rewards - ope_shared_stats.all_reward_model_rewards[
                         range(len(ope_shared_stats.all_actions)), ope_shared_stats.all_actions])) + dm

        return ips, dm, dr



from rl_coach.core_types import TotalStepsCounter, EnvironmentSteps, EnvironmentEpisodes

import pytest


@pytest.mark.unit_test
def test_add_total_steps_counter():
    counter = TotalStepsCounter()
    steps = counter + EnvironmentSteps(10)
    assert steps.num_steps == 10


@pytest.mark.unit_test
def test_add_total_steps_counter_non_zero():
    counter = TotalStepsCounter()
    counter[EnvironmentSteps] += 10
    steps = counter + EnvironmentSteps(10)
    assert steps.num_steps == 20


@pytest.mark.unit_test
def test_total_steps_counter_less_than():
    counter = TotalStepsCounter()
    steps = counter + EnvironmentSteps(0)
    assert not (counter < steps)
    steps = counter + EnvironmentSteps(1)
    assert counter < steps



#
# Copyright (c) 2017 Intel Corporation 
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List

import numpy as np

from rl_coach.core_types import RunPhase, ActionType
from rl_coach.exploration_policies.exploration_policy import ExplorationPolicy, ExplorationParameters
from rl_coach.spaces import ActionSpace


class CategoricalParameters(ExplorationParameters):
    @property
    def path(self):
        return 'rl_coach.exploration_policies.categorical:Categorical'


class Categorical(ExplorationPolicy):
    """
    Categorical exploration policy is intended for discrete action spaces. It expects the action values to
    represent a probability distribution over the action, from which a single action will be sampled.
    In evaluation, the action that has the highest probability will be selected. This is particularly useful for
    actor-critic schemes, where the actors output is a probability distribution over the actions.
    """
    def __init__(self, action_space: ActionSpace):
        """
        :param action_space: the action space used by the environment
        """
        super().__init__(action_space)

    def get_action(self, action_values: List[ActionType]) -> ActionType:
        if self.phase == RunPhase.TRAIN:
            # choose actions according to the probabilities
            return np.random.choice(self.action_space.actions, p=action_values)
        else:
            # take the action with the highest probability
            return np.argmax(action_values)

    def get_control_param(self):
        return 0



#
# Copyright (c) 2017 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import random
from enum import Enum
from itertools import product
from typing import Union, List, Dict, Tuple, Callable

import numpy as np
import scipy
import scipy.spatial

from rl_coach.core_types import ActionType, ActionInfo
from rl_coach.utils import eps


class Space(object):
    """
    A space defines a set of valid values
    """
    def __init__(self, shape: Union[int, tuple, list, np.ndarray], low: Union[None, int, float, np.ndarray]=-np.inf,
                 high: Union[None, int, float, np.ndarray]=np.inf):
        """
        :param shape: the shape of the space
        :param low: the lowest values possible in the space. can be an array defining the lowest values per point,
                    or a single value defining the general lowest values
        :param high: the highest values possible in the space. can be an array defining the highest values per point,
                    or a single value defining the general highest values
        """

        # the number of dimensions is the number of axes in the shape. it will be set in the shape setter
        self.num_dimensions = 0

        # the number of elements is the number of possible actions if the action space was discrete.
        # it will be set in the shape setter
        self.num_elements = 0

        self._low = self._high = None
        self._shape = self.shape = shape
        self._low = self.low = low
        self._high = self.high = high

        # we allow zero sized spaces which means that the space is empty. this is useful for environments with no
        # measurements for example.
        if type(shape) == int and shape < 0:
            raise ValueError("The shape of the space must be a non-negative number")

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, val: Union[int, tuple, list, np.ndarray]):
        # convert the shape to an np.ndarray
        self._shape = val
        if type(self._shape) == int:
            self._shape = np.array([self._shape])
        if type(self._shape) == tuple or type(self._shape) == list:
            self._shape = np.array(self._shape)

        # the shape is now an np.ndarray
        self.num_dimensions = len(self._shape)
        self.num_elements = int(np.prod(self._shape))

    @property
    def low(self):
        if hasattr(self, '_low'):
            return self._low
        else:
            return None

    @low.setter
    def low(self, val: Union[None, int, float, np.ndarray]):
        if type(val) == np.ndarray and type(self.shape) == np.ndarray and np.all(val.shape != self.shape):
            raise ValueError("The low values shape don't match the shape of the space")
        elif self.high is not None and not np.all(self.high >= val):
            raise ValueError("At least one of the axes-parallel lines defining the space has high values which "
                             "are lower than the given low values")
        else:
            self._low = val
            # we allow using a number to define the low values, but we immediately convert it to an array which defines
            # the low values for all the space dimensions in order to expose a consistent value type
            if type(self._low) == int or type(self._low) == float:
                self._low = np.ones(self.shape)*self._low

    @property
    def high(self):
        if hasattr(self, '_high'):
            return self._high
        else:
            return None

    @high.setter
    def high(self, val: Union[None, int, float, np.ndarray]):
        if type(val) == np.ndarray and type(self.shape) == np.ndarray and np.all(val.shape != self.shape):
            raise ValueError("The high values shape don't match the shape of the space")
        elif self.low is not None and not np.all(self.low <= val):
            raise ValueError("At least one of the axes-parallel lines defining the space has low values which "
                             "are higher than the given high values")
        else:
            self._high = val
            # we allow using a number to define the high values, but we immediately convert it to an array which defines
            # the high values for all the space dimensions in order to expose a consistent value type
            if type(self._high) == int or type(self._high) == float:
                self._high = np.ones(self.shape)*self._high

    def contains(self, val: Union[int, float, np.ndarray]) -> bool:
        """
        Checks if value is contained by this space. The shape must match and
        all of the values must be within the low and high bounds.

        :param val: a value to check
        :return: True / False depending on if the val matches the space definition
        """
        if (type(val) == int or type(val) == float) and not np.all(self.shape == np.ones(1)):
            return False
        if type(val) == np.ndarray and not np.all(val.shape == self.shape):
            return False
        if (self.low is not None and not np.all(val >= self.low)) \
                or (self.high is not None and not np.all(val <= self.high)):
            # TODO: check the performance overhead this causes
            return False
        return True

    def is_valid_index(self, index: np.ndarray) -> bool:
        """
        Checks if a given multidimensional index is within the bounds of the shape of the space

        :param index: a multidimensional index
        :return: True if the index is within the shape of the space. False otherwise
        """
        if len(index) != self.num_dimensions:
            return False
        if np.any(index < np.zeros(self.num_dimensions)) or np.any(index >= self.shape):
            return False
        return True

    def sample(self) -> np.ndarray:
        """
        Sample the defined space, either uniformly, if space bounds are defined, or Normal distributed if no
        bounds are defined

        :return: A numpy array sampled from the space
        """
        # if there are infinite bounds, we sample using gaussian noise with mean 0 and std 1
        if np.any(self.low == -np.inf) or np.any(self.high == np.inf):
            return np.random.normal(0, 1, self.shape)
        else:
            return np.random.uniform(self.low, self.high, self.shape)

    def val_matches_space_definition(self, val: Union[int, float, np.ndarray]) -> bool:
        screen.warning(
            "Space.val_matches_space_definition will be deprecated soon. Use "
            "contains instead."
        )
        return self.contains(val)

    def is_point_in_space_shape(self, point: np.ndarray) -> bool:
        screen.warning(
            "Space.is_point_in_space_shape will be deprecated soon. Use "
            "is_valid_index instead."
        )
        return self.is_valid_index(point)


class RewardSpace(Space):
    def __init__(self, shape: Union[int, np.ndarray], low: Union[None, int, float, np.ndarray]=-np.inf,
                 high: Union[None, int, float, np.ndarray]=np.inf,
                 reward_success_threshold: Union[None, int, float]=None):
        super().__init__(shape, low, high)
        self.reward_success_threshold = reward_success_threshold


"""
Observation Spaces
"""


class ObservationSpace(Space):
    def __init__(self, shape: Union[int, np.ndarray], low: Union[None, int, float, np.ndarray]=-np.inf,
                 high: Union[None, int, float, np.ndarray]=np.inf):
        super().__init__(shape, low, high)


class VectorObservationSpace(ObservationSpace):
    """
    An observation space which is defined as a vector of elements. This can be particularly useful for environments
    which return measurements, such as in robotic environments.
    """
    def __init__(self, shape: int, low: Union[None, int, float, np.ndarray]=-np.inf,
                 high: Union[None, int, float, np.ndarray]=np.inf, measurements_names: List[str]=None):
        if measurements_names is None:
            measurements_names = []
        if len(measurements_names) > shape:
            raise ValueError("measurement_names size {} is larger than shape {}.".format(
                len(measurements_names), shape))

        self.measurements_names = measurements_names
        super().__init__(shape, low, high)


class TensorObservationSpace(ObservationSpace):
    """
    An observation space which defines observations with arbitrary shape. This can be particularly useful for
    environments with non image input.
    """
    def __init__(self, shape: np.ndarray, low: -np.inf,
                 high: np.inf):
        super().__init__(shape, low, high)


class PlanarMapsObservationSpace(ObservationSpace):
    """
    An observation space which defines a stack of 2D observations. For example, an environment which returns
    a stack of segmentation maps like in Starcraft.
    """
    def __init__(self, shape: Union[np.ndarray], low: int, high: int, channels_axis: int=-1):
        super().__init__(shape, low, high)
        self.channels_axis = channels_axis

        if not 2 <= len(shape) <= 3:
            raise ValueError("Planar maps observations must have 3 dimensions - a channels dimension and 2 maps "
                             "dimensions, not {}".format(len(shape)))
        if len(shape) == 2:
            self.channels = 1
        else:
            self.channels = shape[channels_axis]


class ImageObservationSpace(PlanarMapsObservationSpace):
    """
    An observation space which is a private case of the PlanarMapsObservationSpace, where the stack of 2D observations
    represent a RGB image, or a grayscale image.
    """
    def __init__(self, shape: Union[np.ndarray], high: int, channels_axis: int=-1):
        # TODO: consider allowing arbitrary low values for images
        super().__init__(shape, 0, high, channels_axis)
        self.has_colors = self.channels == 3
        if not self.channels == 3 and not self.channels == 1:
            raise ValueError("Image observations must have 1 or 3 channels, not {}".format(self.channels))


# TODO: mixed observation spaces (image + measurements, image + segmentation + depth map, etc.)
class StateSpace(object):
    def __init__(self, sub_spaces: Dict[str, Space]):
        self.sub_spaces = sub_spaces

    def __getitem__(self, item):
        return self.sub_spaces[item]

    def __setitem__(self, key, value):
        self.sub_spaces[key] = value


"""
Action Spaces
"""


class ActionSpace(Space):
    def __init__(self, shape: Union[int, np.ndarray], low: Union[None, int, float, np.ndarray]=-np.inf,
                 high: Union[None, int, float, np.ndarray]=np.inf, descriptions: Union[None, List, Dict]=None,
                 default_action: ActionType=None):
        super().__init__(shape, low, high)
        # we allow a mismatch between the number of descriptions and the number of actions.
        # in this case the descriptions for the actions that were not given will be the action index
        if descriptions is not None:
            self.descriptions = descriptions
        else:
            self.descriptions = {}
        self.default_action = default_action

    @property
    def actions(self) -> List[ActionType]:
        raise NotImplementedError("The action space does not have an explicit actions list")

    def sample_with_info(self) -> ActionInfo:
        """
        Get a random action with additional "fake" info

        :return: An action info instance
        """
        return ActionInfo(self.sample())

    def clip_action_to_space(self, action: ActionType) -> ActionType:
        """
        Given an action, clip its values to fit to the action space ranges

        :param action: a given action
        :return: the clipped action
        """
        return action

    def get_description(self, action: np.ndarray) -> str:
        raise NotImplementedError("")

    def __str__(self):
        return "{}: shape = {}, low = {}, high = {}".format(self.__class__.__name__, self.shape, self.low, self.high)

    def __repr__(self):
        return self.__str__()


class AttentionActionSpace(ActionSpace):
    """
    A box selection continuous action space, meaning that the actions are defined as selecting a multidimensional box
    from a given range.
    The actions will be in the form:
    [[low_x, low_y, ...], [high_x, high_y, ...]]
    """
    def __init__(self, shape: int, low: Union[None, int, float, np.ndarray]=-np.inf,
                 high: Union[None, int, float, np.ndarray]=np.inf, descriptions: Union[None, List, Dict]=None,
                 default_action: np.ndarray = None, forced_attention_size: Union[None, int, float, np.ndarray]=None):
        super().__init__(shape, low, high, descriptions)

        self.forced_attention_size = forced_attention_size
        if isinstance(self.forced_attention_size, int) or isinstance(self.forced_attention_size, float):
            self.forced_attention_size = np.ones(self.shape) * self.forced_attention_size

        if self.forced_attention_size is not None and np.all(self.forced_attention_size > (self.high - self.low)):
            raise ValueError("The forced attention size is larger than the action space")

        # default action
        if default_action is None:
            if self.forced_attention_size is not None:
                self.default_action = [self.low*np.ones(self.shape),
                                       (self.low+self.forced_attention_size)*np.ones(self.shape)]
            else:
                self.default_action = [self.low*np.ones(self.shape), self.high*np.ones(self.shape)]
        else:
            self.default_action = default_action

    def sample(self) -> List:
        if self.forced_attention_size is not None:
            sampled_low = np.random.uniform(self.low, self.high-self.forced_attention_size, self.shape)
            sampled_high = sampled_low + self.forced_attention_size
        else:
            sampled_low = np.random.uniform(self.low, self.high, self.shape)
            sampled_high = np.random.uniform(sampled_low, self.high, self.shape)
        return [sampled_low, sampled_high]

    def clip_action_to_space(self, action: ActionType) -> ActionType:
        action = [np.clip(action[0], self.low, self.high), np.clip(action[1], self.low, self.high)]
        return action


class BoxActionSpace(ActionSpace):
    """
    A multidimensional bounded or unbounded continuous action space
    """
    def __init__(self, shape: Union[int, np.ndarray], low: Union[None, int, float, np.ndarray]=-np.inf,
                 high: Union[None, int, float, np.ndarray]=np.inf, descriptions: Union[None, List, Dict]=None,
                 default_action: np.ndarray=None):
        super().__init__(shape, low, high, descriptions)
        self.max_abs_range = np.maximum(np.abs(self.low), np.abs(self.high))

        # default action
        if default_action is None:
            if np.any(np.isinf(self.low)) or np.any(np.isinf(self.high)):
                self.default_action = np.zeros(shape)
            else:
                self.default_action = self.low + (self.high - self.low) / 2
        else:
            self.default_action = default_action

    def clip_action_to_space(self, action: ActionType) -> ActionType:
        action = np.clip(action, self.low, self.high)
        return action


class DiscreteActionSpace(ActionSpace):
    """
    A discrete action space with action indices as actions
    """
    def __init__(self, num_actions: int, descriptions: Union[None, List, Dict]=None, default_action: np.ndarray=None):
        super().__init__(1, low=0, high=num_actions-1, descriptions=descriptions)
        # the number of actions is mapped to high

        # default action
        if default_action is None:
            self.default_action = 0
        else:
            self.default_action = default_action

    @property
    def actions(self) -> List[ActionType]:
        return list(range(0, int(self.high[0]) + 1))

    def sample(self) -> int:
        return np.random.choice(self.actions)

    def sample_with_info(self) -> ActionInfo:
        return ActionInfo(self.sample(),
                          all_action_probabilities=np.full(len(self.actions), 1. / (self.high[0] - self.low[0] + 1)))

    def get_description(self, action: int) -> str:
        if type(self.descriptions) == list and 0 <= action < len(self.descriptions):
            return self.descriptions[action]
        elif type(self.descriptions) == dict and action in self.descriptions.keys():
            return self.descriptions[action]
        elif 0 <= action < self.shape:
            return str(action)
        else:
            raise ValueError("The given action is outside of the action space")


class MultiSelectActionSpace(ActionSpace):
    """
    A discrete action space where multiple actions can be selected at once. The actions are encoded as multi-hot vectors
    """
    def __init__(self, size: int, max_simultaneous_selected_actions: int=1, descriptions: Union[None, List, Dict]=None,
                 default_action: np.ndarray=None, allow_no_action_to_be_selected=True):
        super().__init__(size, low=None, high=None, descriptions=descriptions)
        self.max_simultaneous_selected_actions = max_simultaneous_selected_actions

        if max_simultaneous_selected_actions > size:
            raise ValueError("The maximum simultaneous selected actions can't be larger the max number of actions")

        # create all combinations of actions as a list of actions
        I = [np.eye(size)]*self.max_simultaneous_selected_actions
        self._actions = []
        if allow_no_action_to_be_selected:
            self._actions.append(np.zeros(size))
        self._actions.extend(list(np.unique([np.clip(np.sum(t, axis=0), 0, 1) for t in product(*I)], axis=0)))

        # default action
        if default_action is None:
            self.default_action = self._actions[0]
        else:
            self.default_action = default_action

    @property
    def actions(self) -> List[ActionType]:
        return self._actions

    def sample(self) -> np.ndarray:
        # samples a multi-hot vector
        return random.choice(self.actions)

    def sample_with_info(self) -> ActionInfo:
        return ActionInfo(self.sample(), all_action_probabilities=np.full(len(self.actions), 1. / len(self.actions)))

    def get_description(self, action: np.ndarray) -> str:
        if np.sum(len(np.where(action == 0)[0])) + np.sum(len(np.where(action == 1)[0])) != self.shape or \
                        np.sum(len(np.where(action == 1)[0])) > self.max_simultaneous_selected_actions:
            raise ValueError("The given action is not in the action space")
        selected_actions = np.where(action == 1)[0]
        description = [self.descriptions[a] for a in selected_actions]
        if len(description) == 0:
            description = ['no-op']
        return ' + '.join(description)


class CompoundActionSpace(ActionSpace):
    """
    An action space which consists of multiple sub-action spaces.
    For example, in Starcraft the agent should choose an action identifier from ~550 options (Discrete(550)),
    but it also needs to choose 13 different arguments for the selected action identifier, where each argument is
    by itself an action space. In Starcraft, the arguments are Discrete action spaces as well, but this is not mandatory.
    """
    def __init__(self, sub_spaces: List[ActionSpace]):
        super().__init__(0)
        self.sub_action_spaces = sub_spaces
        # TODO: define the shape, low and high value in a better way

    @property
    def actions(self) -> List[ActionType]:
        return [action_space.actions for action_space in self.sub_action_spaces]

    def sample(self) -> ActionType:
        return [action_space.sample() for action_space in self.sub_action_spaces]

    def clip_action_to_space(self, actions: List[ActionType]) -> ActionType:
        if not isinstance(actions, list) or len(actions) != len(self.sub_action_spaces):
            raise ValueError("The actions to be clipped must be a list with the same number of sub-actions as "
                             "defined in the compound action space.")
        for idx in range(len(self.sub_action_spaces)):
            actions[idx] = self.sub_action_spaces[idx].clip_action_to_space(actions[idx])
        return actions

    def get_description(self, actions: np.ndarray) -> str:
        description = [action_space.get_description(action) for action_space, action in zip(self.sub_action_spaces, actions)]
        return ' + '.join(description)


"""
Goals
"""


class GoalToRewardConversion(object):
    def __init__(self, goal_reaching_reward: float=0):
        self.goal_reaching_reward = goal_reaching_reward

    def convert_distance_to_reward(self, distance: Union[float, np.ndarray]) -> Tuple[float, bool]:
        """
        Given a distance from the goal, return a reward and a flag representing if the goal was reached

        :param distance: the distance from the goal
        :return:
        """
        raise NotImplementedError("")


class ReachingGoal(GoalToRewardConversion):
    """
    get a reward if the goal was reached and 0 otherwise
    """
    def __init__(self, distance_from_goal_threshold: Union[float, np.ndarray], goal_reaching_reward: float=0,
                 default_reward: float=-1):
        """
        :param distance_from_goal_threshold: consider getting to this distance from the goal the same as getting
                                             to the goal
        :param goal_reaching_reward: the reward the agent will get when reaching the goal
        :param default_reward: the reward the agent will get until it reaches the goal
        """
        super().__init__(goal_reaching_reward)
        self.distance_from_goal_threshold = distance_from_goal_threshold
        self.default_reward = default_reward

    def convert_distance_to_reward(self, distance: Union[float, np.ndarray]) -> Tuple[float, bool]:
        if np.all(distance <= self.distance_from_goal_threshold):
            return self.goal_reaching_reward, True
        else:
            return self.default_reward, False


class InverseDistanceFromGoal(GoalToRewardConversion):
    """
    get a reward inversely proportional to the distance from the goal
    """
    def __init__(self, distance_from_goal_threshold: Union[float, np.ndarray], max_reward: float=1):
        """
        :param distance_from_goal_threshold: consider getting to this distance from the goal the same as getting
                                             to the goal
        :param max_reward: the max reward the agent can get
        """
        super().__init__(goal_reaching_reward=max_reward)
        self.distance_from_goal_threshold = distance_from_goal_threshold
        self.max_reward = max_reward

    def convert_distance_to_reward(self, distance: Union[float, np.ndarray]) -> Tuple[float, bool]:
        return min(self.max_reward, 1 / (distance + eps)), distance <= self.distance_from_goal_threshold


class GoalsSpace(VectorObservationSpace, ActionSpace):
    """
    A multidimensional space with a goal type definition. It also behaves as an action space, so that hierarchical
    agents can use it as an output action space.
    The class acts as a wrapper to the target space. So after setting the target space, all the values of the class
    will match the values of the target space (the shape, low, high, etc.)
    """
    class DistanceMetric(Enum):
        Euclidean = 0
        Cosine = 1
        Manhattan = 2

    def __init__(self, goal_name: str, reward_type: GoalToRewardConversion,
                 distance_metric: Union[DistanceMetric, Callable]):
        """
        :param goal_name: the name of the observation space to use as the achieved goal.
        :param reward_type: the reward type to use for converting distances from goal to rewards
        :param distance_metric: the distance metric to use. could be either one of the distances in the
                                DistanceMetric enum, or a custom function that gets two vectors as input and
                                returns the distance between them
        """
        super().__init__(0)
        self.goal_name = goal_name
        self.distance_metric = distance_metric
        self.reward_type = reward_type
        self.target_space = None
        self.max_abs_range = None

    def set_target_space(self, target_space: Space) -> None:
        self.target_space = target_space
        super().__init__(self.target_space.shape, self.target_space.low, self.target_space.high)
        self.max_abs_range = np.maximum(np.abs(self.low), np.abs(self.high))

    def goal_from_state(self, state: Dict):
        """
        Given a state, extract an observation according to the goal_name

        :param state: a dictionary of observations
        :return: the observation corresponding to the goal_name
        """
        return state[self.goal_name]

    def distance_from_goal(self, goal: np.ndarray, state: dict) -> float:
        """
        Given a state, check its distance from the goal

        :param goal: a numpy array representing the goal
        :param state: a dict representing the state
        :return: the distance from the goal
        """
        state_value = self.goal_from_state(state)

        # calculate distance
        if self.distance_metric == self.DistanceMetric.Cosine:
            dist = scipy.spatial.distance.cosine(goal, state_value)
        elif self.distance_metric == self.DistanceMetric.Euclidean:
            dist = scipy.spatial.distance.euclidean(goal, state_value)
        elif self.distance_metric == self.DistanceMetric.Manhattan:
            dist = scipy.spatial.distance.cityblock(goal, state_value)
        elif callable(self.distance_metric):
            dist = self.distance_metric(goal, state_value)
        else:
            raise ValueError("The given distance metric for the goal is not valid.")

        return dist

    def get_reward_for_goal_and_state(self, goal: np.ndarray, state: dict) -> Tuple[float, bool]:
        """
        Given a state, check if the goal was reached and return a reward accordingly

        :param goal: a numpy array representing the goal
        :param state: a dict representing the state
        :return: the reward for the current goal and state pair and a boolean representing if the goal was reached
        """
        dist = self.distance_from_goal(goal, state)
        return self.reward_type.convert_distance_to_reward(dist)


class AgentSelection(DiscreteActionSpace):
    """
    An discrete action space which is bounded by the number of agents to select from
    """
    def __init__(self, num_agents: int):
        super().__init__(num_agents)


class SpacesDefinition(object):
    """
    A container class that allows passing the definitions of all the spaces at once
    """
    def __init__(self,
                 state: StateSpace,
                 goal: Union[ObservationSpace, None],
                 action: ActionSpace,
                 reward: RewardSpace):
        self.state = state
        self.goal = goal
        self.action = action
        self.reward = reward




import argparse
import pytest
import time
from kubernetes import client, config


class EKSHandler():

    def __init__(self, cluster, build_num, test_name, test_command, image, cpu, memory, working_dir):
        self.cluster = cluster
        self.build_num = build_num
        self.test_name = test_name
        self.test_command = test_command
        self.image = image
        self.cpu = cpu
        self.memory = memory
        self.refresh_config()
        self.namespace = '{}-{}'.format(test_name, build_num)
        self.create_namespace()
        self.working_dir = working_dir

    def refresh_config(self):
        # on AWS tokens only last 10 minutes so this must periodically be
        # called to prevent auth related errors
        config.load_kube_config()
        self.corev1_api = client.CoreV1Api()

    def create_namespace(self):
        namespace = client.V1Namespace(
            api_version='v1',
            kind="Namespace",
            metadata=client.V1ObjectMeta(name=self.namespace)
        )

        try:
            self.corev1_api.create_namespace(namespace)
        except client.rest.ApiException as e:
            raise RuntimeError("Failed to create namesapce. Got exception: {}".format(e))

    def deploy(self):
        container = client.V1Container(
            name=self.test_name,
            image=self.image,
            command=['/bin/bash', '-c'],
            args=[self.test_command],
            image_pull_policy='Always',
            working_dir=self.working_dir,
            stdin=True,
            tty=True
        )
        pod_spec = client.V1PodSpec(
            containers=[container],
            restart_policy='Never'
        )
        pod = client.V1Pod(
            api_version="v1",
            kind="Pod",
            metadata=client.V1ObjectMeta(name=self.test_name),
            spec=pod_spec
        )

        try:
            self.corev1_api.create_namespaced_pod(self.namespace, pod)
        except client.rest.ApiException as e:
            print("Got exception: {} while creating a pod".format(e))
            return 1

        return 0

    def print_logs(self):
        while True:
            time.sleep(10)
            # Try to tail the pod logs
            try:
                for line in self.corev1_api.read_namespaced_pod_log(
                    self.test_name, self.namespace, follow=True,
                    _preload_content=False
                ):
                    print(line.decode('utf-8'), flush=True, end='')
                # above call blocks for pod lifetime, so we may need to refresh tokens
                self.refresh_config()

            except client.rest.ApiException as e:
                print("Got exception: {} while reading pod logs".format(e))
                pass

            try:
                pod = self.corev1_api.read_namespaced_pod(self.test_name, self.namespace)
            except client.rest.ApiException as e:
                print("Got exception: {} while reading pod".format(e))
                continue

            if not hasattr(pod, 'status') or not pod.status:
                continue
            if not hasattr(pod.status, 'container_statuses') or not pod.status.container_statuses:
                continue

            for container_status in pod.status.container_statuses:
                if container_status.state.waiting is not None:
                    if container_status.state.waiting.reason == 'Error' or \
                       container_status.state.waiting.reason == 'CrashLoopBackOff' or \
                       container_status.state.waiting.reason == 'ImagePullBackOff' or \
                       container_status.state.waiting.reason == 'ErrImagePull':
                        return
                if container_status.state.terminated is not None:
                    return

    def get_return_status(self):
        # This part will get executed if the pod is one of the following phases: not ready, failed or terminated.
        # Check if the pod has errored out, else just try again.
        # Get the pod
        try:
            pod = self.corev1_api.read_namespaced_pod(self.test_name, self.namespace)
        except client.rest.ApiException as e:
            print("Got exception: {} while reading pod".format(e))
            return 1

        if not hasattr(pod, 'status') or not pod.status:
            return 0
        if not hasattr(pod.status, 'container_statuses') or not pod.status.container_statuses:
            return 0

        for container_status in pod.status.container_statuses:
            if container_status.state.waiting is not None:
                if container_status.state.waiting.reason == 'Error' or \
                   container_status.state.waiting.reason == 'CrashLoopBackOff' or \
                   container_status.state.waiting.reason == 'ImagePullBackOff' or \
                   container_status.state.waiting.reason == 'ErrImagePull':
                    return 1
            if container_status.state.terminated is not None:
                return container_status.state.terminated.exit_code

    def cleanup(self):

        # Delete pod
        try:
            self.corev1_api.delete_namespaced_pod(self.test_name, self.namespace, client.V1DeleteOptions())
        except client.rest.ApiException as e:
            print("Got exception while deleting pod: {}".format(e))

        # Delete namespace
        try:
            self.corev1_api.delete_namespace(self.namespace, client.V1DeleteOptions())
        except client.rest.ApiException as e:
            print("Got exception while deleting namespace: {}".format(e))


if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '-c', '--cluster', help="(string) Name of the cluster", type=str, required=True
    )
    parser.add_argument(
        '-bn', '--build-num', help="(int) CI Build number", type=int, required=True
    )
    parser.add_argument(
        '-tn', '--test-name', help="(string) Name of the test", type=str, required=True
    )
    parser.add_argument(
        '-tc', '--test-command', help="(string) command to execute", type=str, required=True
    )
    parser.add_argument(
        '-i', '--image', help="(string) Container image", type=str, required=True
    )
    parser.add_argument(
        '-cpu', help="(string) Units of cpu to use", type=str, required=True
    )
    parser.add_argument(
        '-mem', help="(string) The amount in megabytes", type=str, required=True
    )
    parser.add_argument(
        '--working-dir', help="(string) The working dir in the container", type=str, required=False,
        default='/root/src/docker'
    )
    args = parser.parse_args()

    obj = EKSHandler(
        args.cluster, args.build_num, args.test_name, args.test_command,
        args.image, args.cpu, args.mem, args.working_dir
    )

    if obj.deploy() != 0:
        obj.cleanup()
        pytest.fail("Failed to deploy")

    obj.print_logs()

    if obj.get_return_status() != 0:
        obj.cleanup()
        pytest.fail("Failed to run tests")

    obj.cleanup()



#
# Copyright (c) 2017 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from itertools import product
from typing import Union, List

import numpy as np

from rl_coach.filters.action.partial_discrete_action_space_map import PartialDiscreteActionSpaceMap
from rl_coach.spaces import BoxActionSpace, DiscreteActionSpace


class BoxDiscretization(PartialDiscreteActionSpaceMap):
    """
    Discretizes a continuous action space into a discrete action space, allowing the usage of
    agents such as DQN for continuous environments such as MuJoCo. Given the number of bins to discretize into, the
    original continuous action space is uniformly separated into the given number of bins, each mapped to a discrete
    action index. Each discrete action is mapped to a single N dimensional action in the BoxActionSpace action space.
    For example, if the original actions space is between -1 and 1 and 5 bins were selected, the new action
    space will consist of 5 actions mapped to -1, -0.5, 0, 0.5 and 1.
    """
    def __init__(self, num_bins_per_dimension: Union[int, List[int]], force_int_bins=False):
        """
        :param num_bins_per_dimension: The number of bins to use for each dimension of the target action space.
                                       The bins will be spread out uniformly over this space
        :param force_int_bins: force the bins to represent only integer actions. for example, if the action space is in
                               the range 0-10 and there are 5 bins, then the bins will be placed at 0, 2, 5, 7, 10,
                               instead of 0, 2.5, 5, 7.5, 10.
        """
        # we allow specifying either a single number for all dimensions, or a single number per dimension in the target
        # action space
        self.num_bins_per_dimension = num_bins_per_dimension
        self.force_int_bins = force_int_bins
        super().__init__()

    def validate_output_action_space(self, output_action_space: BoxActionSpace):
        if not isinstance(output_action_space, BoxActionSpace):
            raise ValueError("BoxActionSpace discretization only works with an output space of type BoxActionSpace. "
                             "The given output space is {}".format(output_action_space))

        if len(self.num_bins_per_dimension) != output_action_space.shape:
            # TODO: this check is not sufficient. it does not deal with actions spaces with more than one axis
            raise ValueError("The length of the list of bins per dimension ({}) does not match the number of "
                             "dimensions in the action space ({})"
                             .format(len(self.num_bins_per_dimension), output_action_space))

    def get_unfiltered_action_space(self, output_action_space: BoxActionSpace) -> DiscreteActionSpace:
        if isinstance(self.num_bins_per_dimension, int):
            self.num_bins_per_dimension = np.ones(output_action_space.shape) * self.num_bins_per_dimension

        bins = []
        for i in range(len(output_action_space.low)):
            dim_bins = np.linspace(output_action_space.low[i], output_action_space.high[i],
                                   self.num_bins_per_dimension[i])
            if self.force_int_bins:
                dim_bins = dim_bins.astype(int)
            bins.append(dim_bins)
        self.target_actions = [list(action) for action in list(product(*bins))]

        return super().get_unfiltered_action_space(output_action_space)



#
# Copyright (c) 2017 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np

from rl_coach.core_types import ObservationType
from rl_coach.filters.observation.observation_filter import ObservationFilter
from rl_coach.spaces import ObservationSpace


class ObservationSqueezeFilter(ObservationFilter):
    """
    Removes redundant axes from the observation, which are axes with a dimension of 1.
    """
    def __init__(self, axis: int = None):
        """
        :param axis: Specifies which axis to remove. If set to None, all the axes of size 1 will be removed.
        """
        super().__init__()
        self.axis = axis

    def validate_input_observation_space(self, input_observation_space: ObservationSpace):
        if self.axis is None:
            return

        shape = input_observation_space.shape
        if self.axis >= len(shape) or self.axis < -len(shape):
            raise ValueError("The given axis does not exist in the context of the input observation shape. ")

    def filter(self, observation: ObservationType, update_internal_state: bool=True) -> ObservationType:
        return observation.squeeze(axis=self.axis)

    def get_filtered_observation_space(self, input_observation_space: ObservationSpace) -> ObservationSpace:
        dummy_tensor = np.random.rand(*tuple(input_observation_space.shape))
        input_observation_space.shape = dummy_tensor.squeeze(axis=self.axis).shape
        return input_observation_space



#
# Copyright (c) 2019 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""PyTest configuration."""

import os
import shutil
import sys
import pytest
import rl_coach.tests.utils.args_utils as a_utils
import rl_coach.tests.utils.presets_utils as p_utils
from rl_coach.tests.utils.definitions import Definitions as Def
from os import path


@pytest.fixture(scope="module", params=list(p_utils.collect_presets()))
def preset_name(request):
    """
    Return all preset names
    """
    yield request.param


@pytest.fixture(scope="function", params=list(a_utils.collect_args()))
def flag(request):
    """
    Return flags names in function scope
    """
    yield request.param


@pytest.fixture(scope="module", params=list(a_utils.collect_preset_for_args()))
def preset_args(request):
    """
    Return preset names that can be used for args testing only; working in
    module scope
    """
    yield request.param


@pytest.fixture(scope="module", params=list(a_utils.collect_preset_for_seed()))
def preset_args_for_seed(request):
    """
    Return preset names that can be used for args testing only and for special
    action when using seed argument; working in module scope
    """
    yield request.param


@pytest.fixture(scope="module",
                params=list(a_utils.collect_preset_for_mxnet()))
def preset_for_mxnet_args(request):
    """
    Return preset names that can be used for args testing only; this special
    fixture will be used for mxnet framework only. working in module scope
    """
    yield request.param


@pytest.fixture(scope="function")
def clres(request):
    """
    Create both file csv and log for testing
    :yield: class of both files paths
    """

    class CreateCsvLog:
        """
        Create a test and log paths
        """
        def __init__(self, csv, log, pattern):
            self.exp_path = csv
            self.stdout = log
            self.fn_pattern = pattern

    # get preset name from test request params
    idx = 0 if 'preset' in list(request.node.funcargs.items())[0][0] else 1
    p_name = list(request.node.funcargs.items())[idx][1]

    p_valid_params = p_utils.validation_params(p_name)

    sys.path.append('.')
    test_name = 'ExpName_{}'.format(p_name)
    test_path = os.path.join(Def.Path.experiments, test_name)
    if path.exists(test_path):
        shutil.rmtree(test_path)

    # get the stdout for logs results
    log_file_name = 'test_log_{}.txt'.format(p_name)
    stdout = open(log_file_name, 'w')
    fn_pattern = '*.csv' if p_valid_params.num_workers > 1 else 'worker_0*.csv'

    res = CreateCsvLog(test_path, stdout, fn_pattern)

    yield res

    # clean files
    if path.exists(res.exp_path):
        shutil.rmtree(res.exp_path)

    if path.exists(res.stdout.name):
        os.remove(res.stdout.name)



import os
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))

import pytest
import numpy as np

from rl_coach.filters.observation.observation_to_uint8_filter import ObservationToUInt8Filter
from rl_coach.spaces import ObservationSpace
from rl_coach.core_types import EnvResponse
from rl_coach.filters.filter import InputFilter


@pytest.mark.unit_test
def test_filter():
    # make an RGB observation smaller
    uint8_filter = InputFilter()
    uint8_filter.add_observation_filter('observation', 'to_uint8', ObservationToUInt8Filter(input_low=0, input_high=255))

    observation = np.random.rand(20, 30, 3)*255.0
    env_response = EnvResponse(next_state={'observation': observation}, reward=0, game_over=False)

    result = uint8_filter.filter(env_response)[0]
    unfiltered_observation = env_response.next_state['observation']
    filtered_observation = result.next_state['observation']

    # make sure the original observation is unchanged
    assert unfiltered_observation.dtype == 'float64'

    # make sure the filtering is done correctly
    assert filtered_observation.dtype == 'uint8'
    assert np.all(filtered_observation == observation.astype('uint8'))


@pytest.mark.unit_test
def test_get_filtered_observation_space():
    # error on observation space with values not matching the filter configuration
    uint8_filter = InputFilter()
    uint8_filter.add_observation_filter('observation', 'to_uint8', ObservationToUInt8Filter(input_low=0, input_high=200))

    observation_space = ObservationSpace(np.array([1, 2, 3]), 0, 100)
    with pytest.raises(ValueError):
        uint8_filter.get_filtered_observation_space('observation', observation_space)

    # verify output observation space is correct
    observation_space = ObservationSpace(np.array([1, 2, 3]), 0, 200)
    result = uint8_filter.get_filtered_observation_space('observation', observation_space)
    assert np.all(result.high == 255)
    assert np.all(result.low == 0)
    assert np.all(result.shape == observation_space.shape)



#
# Copyright (c) 2017 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List, Union

from rl_coach.base_parameters import EmbedderScheme, NetworkComponentParameters

MOD_NAMES = {'image': 'ImageEmbedder', 'vector': 'VectorEmbedder', 'tensor': 'TensorEmbedder'}

class InputEmbedderParameters(NetworkComponentParameters):
    def __init__(self, activation_function: str='relu', scheme: Union[List, EmbedderScheme]=EmbedderScheme.Medium,
                 batchnorm: bool=False, dropout_rate: float=0.0, name: str='embedder', input_rescaling=None,
                 input_offset=None, input_clipping=None, dense_layer=None, is_training=False):
        super().__init__(dense_layer=dense_layer)
        self.activation_function = activation_function
        self.scheme = scheme
        self.batchnorm = batchnorm
        self.dropout_rate = dropout_rate

        if input_rescaling is None:
            input_rescaling = {'image': 255.0, 'vector': 1.0, 'tensor': 1.0}
        if input_offset is None:
            input_offset = {'image': 0.0, 'vector': 0.0, 'tensor': 0.0}

        self.input_rescaling = input_rescaling
        self.input_offset = input_offset
        self.input_clipping = input_clipping
        self.name = name
        self.is_training = is_training

    def path(self, emb_type):
        return 'rl_coach.architectures.tensorflow_components.embedders:' + MOD_NAMES[emb_type]



#
# Copyright (c) 2017 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from rl_coach.data_stores.data_store import DataStore, DataStoreParameters
from minio import Minio
from minio.error import ResponseError
from configparser import ConfigParser, Error
from rl_coach.checkpoint import CheckpointStateFile
from rl_coach.data_stores.data_store import SyncFiles

import os
import time
import io


class S3DataStoreParameters(DataStoreParameters):
    def __init__(self, ds_params, creds_file: str = None, end_point: str = None, bucket_name: str = None,
                 checkpoint_dir: str = None, expt_dir: str = None):

        super().__init__(ds_params.store_type, ds_params.orchestrator_type, ds_params.orchestrator_params)
        self.creds_file = creds_file
        self.end_point = end_point
        self.bucket_name = bucket_name
        self.checkpoint_dir = checkpoint_dir
        self.expt_dir = expt_dir


class S3DataStore(DataStore):
    """
    An implementation of the data store using S3 for storing policy checkpoints when using Coach in distributed mode.
    The policy checkpoints are written by the trainer and read by the rollout worker.
    """

    def __init__(self, params: S3DataStoreParameters):
        """
        :param params: The parameters required to use the S3 data store.
        """

        super(S3DataStore, self).__init__(params)
        self.params = params
        access_key = None
        secret_key = None
        if params.creds_file:
            config = ConfigParser()
            config.read(params.creds_file)
            try:
                access_key = config.get('default', 'aws_access_key_id')
                secret_key = config.get('default', 'aws_secret_access_key')
            except Error as e:
                print("Error when reading S3 credentials file: %s", e)
        else:
            access_key = os.environ.get('ACCESS_KEY_ID')
            secret_key = os.environ.get('SECRET_ACCESS_KEY')
        self.mc = Minio(self.params.end_point, access_key=access_key, secret_key=secret_key)

    def deploy(self) -> bool:
        return True

    def get_info(self):
        return "s3://{}/{}".format(self.params.bucket_name)

    def undeploy(self) -> bool:
        return True

    def save_to_store(self):
        """
        save_to_store() uploads the policy checkpoint, gifs and videos to the S3 data store. It reads the checkpoint state files and
        uploads only the latest checkpoint files to S3. It is used by the trainer in Coach when used in the distributed mode.
        """
        try:
            # remove lock file if it exists
            self.mc.remove_object(self.params.bucket_name, SyncFiles.LOCKFILE.value)

            # Acquire lock
            self.mc.put_object(self.params.bucket_name, SyncFiles.LOCKFILE.value, io.BytesIO(b''), 0)

            state_file = CheckpointStateFile(os.path.abspath(self.params.checkpoint_dir))
            if state_file.exists():
                ckpt_state = state_file.read()
                checkpoint_file = None
                for root, dirs, files in os.walk(self.params.checkpoint_dir):
                    for filename in files:
                        if filename == CheckpointStateFile.checkpoint_state_filename:
                            checkpoint_file = (root, filename)
                            continue
                        if filename.startswith(ckpt_state.name):
                            abs_name = os.path.abspath(os.path.join(root, filename))
                            rel_name = os.path.relpath(abs_name, self.params.checkpoint_dir)
                            self.mc.fput_object(self.params.bucket_name, rel_name, abs_name)

                abs_name = os.path.abspath(os.path.join(checkpoint_file[0], checkpoint_file[1]))
                rel_name = os.path.relpath(abs_name, self.params.checkpoint_dir)
                self.mc.fput_object(self.params.bucket_name, rel_name, abs_name)

            # release lock
            self.mc.remove_object(self.params.bucket_name, SyncFiles.LOCKFILE.value)

            if self.params.expt_dir and os.path.exists(self.params.expt_dir):
                for filename in os.listdir(self.params.expt_dir):
                    if filename.endswith((".csv", ".json")):
                        self.mc.fput_object(self.params.bucket_name, filename, os.path.join(self.params.expt_dir, filename))

            if self.params.expt_dir and os.path.exists(os.path.join(self.params.expt_dir, 'videos')):
                for filename in os.listdir(os.path.join(self.params.expt_dir, 'videos')):
                        self.mc.fput_object(self.params.bucket_name, filename, os.path.join(self.params.expt_dir, 'videos', filename))

            if self.params.expt_dir and os.path.exists(os.path.join(self.params.expt_dir, 'gifs')):
                for filename in os.listdir(os.path.join(self.params.expt_dir, 'gifs')):
                        self.mc.fput_object(self.params.bucket_name, filename, os.path.join(self.params.expt_dir, 'gifs', filename))
        except ResponseError as e:
            print("Got exception: %s\n while saving to S3", e)

    def load_from_store(self):
        """
        load_from_store() downloads a new checkpoint from the S3 data store when it is not available locally. It is used
        by the rollout workers when using Coach in distributed mode.
        """
        try:
            state_file = CheckpointStateFile(os.path.abspath(self.params.checkpoint_dir))

            # wait until lock is removed
            while True:
                objects = self.mc.list_objects_v2(self.params.bucket_name, SyncFiles.LOCKFILE.value)

                if next(objects, None) is None:
                    try:
                        # fetch checkpoint state file from S3
                        self.mc.fget_object(self.params.bucket_name, state_file.filename, state_file.path)
                    except Exception as e:
                        continue
                    break
                time.sleep(10)

            # Check if there's a finished file
            objects = self.mc.list_objects_v2(self.params.bucket_name, SyncFiles.FINISHED.value)

            if next(objects, None) is not None:
                try:
                    self.mc.fget_object(
                        self.params.bucket_name, SyncFiles.FINISHED.value,
                        os.path.abspath(os.path.join(self.params.checkpoint_dir, SyncFiles.FINISHED.value))
                    )
                except Exception as e:
                    pass

            checkpoint_state = state_file.read()
            if checkpoint_state is not None:
                objects = self.mc.list_objects_v2(self.params.bucket_name, prefix=checkpoint_state.name, recursive=True)
                for obj in objects:
                    filename = os.path.abspath(os.path.join(self.params.checkpoint_dir, obj.object_name))
                    if not os.path.exists(filename):
                        self.mc.fget_object(obj.bucket_name, obj.object_name, filename)

        except ResponseError as e:
            print("Got exception: %s\n while loading from S3", e)



import mxnet as mx
import os
import pytest
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))


from rl_coach.architectures.mxnet_components.heads.q_head import QHead, QHeadLoss
from rl_coach.agents.clipped_ppo_agent import ClippedPPOAgentParameters
from rl_coach.spaces import SpacesDefinition, DiscreteActionSpace



@pytest.mark.unit_test
def test_q_head_loss():
    loss_fn = QHeadLoss()
    # example with batch_size of 3, and num_actions of 2
    target_q_values = mx.nd.array(((3, 5), (-1, -2), (0, 2)))
    pred_q_values_worse = mx.nd.array(((6, 5), (-1, -2), (0, 2)))
    pred_q_values_better = mx.nd.array(((4, 5), (-2, -2), (1, 2)))
    loss_worse = loss_fn(pred_q_values_worse, target_q_values)
    loss_better = loss_fn(pred_q_values_better, target_q_values)
    assert len(loss_worse) == 1  # (LOSS)
    loss_worse_val = loss_worse[0]
    assert loss_worse_val.ndim == 1
    assert loss_worse_val.shape[0] == 1
    assert len(loss_better) == 1  # (LOSS)
    loss_better_val = loss_better[0]
    assert loss_better_val.ndim == 1
    assert loss_better_val.shape[0] == 1
    assert loss_worse_val > loss_better_val


@pytest.mark.unit_test
def test_v_head_loss_weight():
    target_q_values = mx.nd.array(((3, 5), (-1, -2), (0, 2)))
    pred_q_values = mx.nd.array(((4, 5), (-2, -2), (1, 2)))
    loss_fn = QHeadLoss()
    loss = loss_fn(pred_q_values, target_q_values)
    loss_fn_weighted = QHeadLoss(weight=0.5)
    loss_weighted = loss_fn_weighted(pred_q_values, target_q_values)
    assert loss[0] == loss_weighted[0]*2


@pytest.mark.unit_test
def test_ppo_v_head():
    agent_parameters = ClippedPPOAgentParameters()
    num_actions = 5
    action_space = DiscreteActionSpace(num_actions=num_actions)
    spaces = SpacesDefinition(state=None, goal=None, action=action_space, reward=None)
    value_net = QHead(agent_parameters=agent_parameters,
                      spaces=spaces,
                      network_name="test_q_head")
    value_net.initialize()
    batch_size = 15
    middleware_data = mx.nd.random.uniform(shape=(batch_size, 100))
    values = value_net(middleware_data)
    assert values.ndim == 2  # (batch_size, num_actions)
    assert values.shape[0] == batch_size
    assert values.shape[1] == num_actions


#
# Copyright (c) 2017 Intel Corporation 
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import copy
from typing import Union
from collections import OrderedDict

import numpy as np

from rl_coach.agents.actor_critic_agent import ActorCriticAgent
from rl_coach.agents.agent import Agent
from rl_coach.architectures.embedder_parameters import InputEmbedderParameters
from rl_coach.architectures.head_parameters import DDPGActorHeadParameters, VHeadParameters
from rl_coach.architectures.middleware_parameters import FCMiddlewareParameters
from rl_coach.base_parameters import NetworkParameters, AlgorithmParameters, \
    AgentParameters, EmbedderScheme
from rl_coach.core_types import ActionInfo, EnvironmentSteps
from rl_coach.exploration_policies.ou_process import OUProcessParameters
from rl_coach.memories.episodic.episodic_experience_replay import EpisodicExperienceReplayParameters
from rl_coach.spaces import BoxActionSpace, GoalsSpace


class DDPGCriticNetworkParameters(NetworkParameters):
    def __init__(self):
        super().__init__()
        self.input_embedders_parameters = {'observation': InputEmbedderParameters(batchnorm=True),
                                            'action': InputEmbedderParameters(scheme=EmbedderScheme.Shallow)}
        self.middleware_parameters = FCMiddlewareParameters()
        self.heads_parameters = [VHeadParameters()]
        self.optimizer_type = 'Adam'
        self.batch_size = 64
        self.async_training = False
        self.learning_rate = 0.001
        self.create_target_network = True
        self.shared_optimizer = True
        self.scale_down_gradients_by_number_of_workers_for_sync_training = False


class DDPGActorNetworkParameters(NetworkParameters):
    def __init__(self):
        super().__init__()
        self.input_embedders_parameters = {'observation': InputEmbedderParameters(batchnorm=True)}
        self.middleware_parameters = FCMiddlewareParameters(batchnorm=True)
        self.heads_parameters = [DDPGActorHeadParameters()]
        self.optimizer_type = 'Adam'
        self.batch_size = 64
        self.async_training = False
        self.learning_rate = 0.0001
        self.create_target_network = True
        self.shared_optimizer = True
        self.scale_down_gradients_by_number_of_workers_for_sync_training = False


class DDPGAlgorithmParameters(AlgorithmParameters):
    """
    :param num_steps_between_copying_online_weights_to_target: (StepMethod)
        The number of steps between copying the online network weights to the target network weights.

    :param rate_for_copying_weights_to_target: (float)
        When copying the online network weights to the target network weights, a soft update will be used, which
        weight the new online network weights by rate_for_copying_weights_to_target

    :param num_consecutive_playing_steps: (StepMethod)
        The number of consecutive steps to act between every two training iterations

    :param use_target_network_for_evaluation: (bool)
        If set to True, the target network will be used for predicting the actions when choosing actions to act.
        Since the target network weights change more slowly, the predicted actions will be more consistent.

    :param action_penalty: (float)
        The amount by which to penalize the network on high action feature (pre-activation) values.
        This can prevent the actions features from saturating the TanH activation function, and therefore prevent the
        gradients from becoming very low.

    :param clip_critic_targets: (Tuple[float, float] or None)
        The range to clip the critic target to in order to prevent overestimation of the action values.

    :param use_non_zero_discount_for_terminal_states: (bool)
        If set to True, the discount factor will be used for terminal states to bootstrap the next predicted state
        values. If set to False, the terminal states reward will be taken as the target return for the network.
    """
    def __init__(self):
        super().__init__()
        self.num_steps_between_copying_online_weights_to_target = EnvironmentSteps(1)
        self.rate_for_copying_weights_to_target = 0.001
        self.num_consecutive_playing_steps = EnvironmentSteps(1)
        self.use_target_network_for_evaluation = False
        self.action_penalty = 0
        self.clip_critic_targets = None  # expected to be a tuple of the form (min_clip_value, max_clip_value) or None
        self.use_non_zero_discount_for_terminal_states = False


class DDPGAgentParameters(AgentParameters):
    def __init__(self):
        super().__init__(algorithm=DDPGAlgorithmParameters(),
                         exploration=OUProcessParameters(),
                         memory=EpisodicExperienceReplayParameters(),
                         networks=OrderedDict([("actor", DDPGActorNetworkParameters()),
                                               ("critic", DDPGCriticNetworkParameters())]))

    @property
    def path(self):
        return 'rl_coach.agents.ddpg_agent:DDPGAgent'


# Deep Deterministic Policy Gradients Network - https://arxiv.org/pdf/1509.02971.pdf
class DDPGAgent(ActorCriticAgent):
    def __init__(self, agent_parameters, parent: Union['LevelManager', 'CompositeAgent']=None):
        super().__init__(agent_parameters, parent)

        self.q_values = self.register_signal("Q")
        self.TD_targets_signal = self.register_signal("TD targets")
        self.action_signal = self.register_signal("actions")

    def learn_from_batch(self, batch):
        actor = self.networks['actor']
        critic = self.networks['critic']

        actor_keys = self.ap.network_wrappers['actor'].input_embedders_parameters.keys()
        critic_keys = self.ap.network_wrappers['critic'].input_embedders_parameters.keys()

        # TD error = r + discount*max(q_st_plus_1) - q_st
        next_actions, actions_mean = actor.parallel_prediction([
            (actor.target_network, batch.next_states(actor_keys)),
            (actor.online_network, batch.states(actor_keys))
        ])

        critic_inputs = copy.copy(batch.next_states(critic_keys))
        critic_inputs['action'] = next_actions
        q_st_plus_1 = critic.target_network.predict(critic_inputs)

        # calculate the bootstrapped TD targets while discounting terminal states according to
        # use_non_zero_discount_for_terminal_states
        if self.ap.algorithm.use_non_zero_discount_for_terminal_states:
            TD_targets = batch.rewards(expand_dims=True) + self.ap.algorithm.discount * q_st_plus_1
        else:
            TD_targets = batch.rewards(expand_dims=True) + \
                         (1.0 - batch.game_overs(expand_dims=True)) * self.ap.algorithm.discount * q_st_plus_1

        # clip the TD targets to prevent overestimation errors
        if self.ap.algorithm.clip_critic_targets:
            TD_targets = np.clip(TD_targets, *self.ap.algorithm.clip_critic_targets)

        self.TD_targets_signal.add_sample(TD_targets)

        # get the gradients of the critic output with respect to the action
        critic_inputs = copy.copy(batch.states(critic_keys))
        critic_inputs['action'] = actions_mean
        action_gradients = critic.online_network.predict(critic_inputs,
                                                         outputs=critic.online_network.gradients_wrt_inputs[0]['action'])

        # train the critic
        critic_inputs = copy.copy(batch.states(critic_keys))
        critic_inputs['action'] = batch.actions(len(batch.actions().shape) == 1)
        result = critic.train_and_sync_networks(critic_inputs, TD_targets)
        total_loss, losses, unclipped_grads = result[:3]

        # apply the gradients from the critic to the actor
        initial_feed_dict = {actor.online_network.gradients_weights_ph[0]: -action_gradients}
        gradients = actor.online_network.predict(batch.states(actor_keys),
                                                 outputs=actor.online_network.weighted_gradients[0],
                                                 initial_feed_dict=initial_feed_dict)

        if actor.has_global:
            actor.apply_gradients_to_global_network(gradients)
            actor.update_online_network()
        else:
            actor.apply_gradients_to_online_network(gradients)

        return total_loss, losses, unclipped_grads

    def train(self):
        return Agent.train(self)

    def choose_action(self, curr_state):
        if not (isinstance(self.spaces.action, BoxActionSpace) or isinstance(self.spaces.action, GoalsSpace)):
            raise ValueError("DDPG works only for continuous control problems")
        # convert to batch so we can run it through the network
        tf_input_state = self.prepare_batch_for_inference(curr_state, 'actor')
        if self.ap.algorithm.use_target_network_for_evaluation:
            actor_network = self.networks['actor'].target_network
        else:
            actor_network = self.networks['actor'].online_network

        action_values = actor_network.predict(tf_input_state).squeeze()

        action = self.exploration_policy.get_action(action_values)

        self.action_signal.add_sample(action)

        # get q value
        tf_input_state = self.prepare_batch_for_inference(curr_state, 'critic')
        action_batch = np.expand_dims(action, 0)
        if type(action) != np.ndarray:
            action_batch = np.array([[action]])
        tf_input_state['action'] = action_batch
        q_value = self.networks['critic'].online_network.predict(tf_input_state)[0]
        self.q_values.add_sample(q_value)

        action_info = ActionInfo(action=action,
                                 action_value=q_value)

        return action_info



#
# Copyright (c) 2017 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import copy
import os
from collections import OrderedDict
from copy import deepcopy
from typing import Dict, Union, List

from rl_coach.core_types import EnvResponse, ActionInfo, Transition
from rl_coach.spaces import ActionSpace, RewardSpace, ObservationSpace
from rl_coach.utils import force_list


class Filter(object):
    def __init__(self, name=None):
        self.name = name

    def reset(self) -> None:
        """
        Called from reset() and implements the reset logic for the filter.
        :param name: the filter's name
        :return: None
        """
        pass

    def filter(self, env_response: Union[EnvResponse, Transition], update_internal_state: bool=True) \
            -> Union[EnvResponse, Transition]:
        """
        Filter some values in the env and return the filtered env_response
        This is the function that each filter should update
        :param update_internal_state: should the filter's internal state change due to this call
        :param env_response: the input env_response
        :return: the filtered env_response
        """
        raise NotImplementedError("")

    def set_device(self, device, memory_backend_params=None, mode='numpy') -> None:
        """
        An optional function that allows the filter to get the device if it is required to use tensorflow ops
        :param device: the device to use
        :param memory_backend_params: parameters associated with the memory backend
        :param mode: arithmetic backend to be used {numpy | tf}
        :return: None
        """
        pass

    def set_session(self, sess) -> None:
        """
        An optional function that allows the filter to get the session if it is required to use tensorflow ops
        :param sess: the session
        :return: None
        """
        pass

    def save_state_to_checkpoint(self, checkpoint_dir, checkpoint_prefix)->None:
        """
        Save the filter's internal state to a checkpoint to file, so that it can be later restored.
        :param checkpoint_dir: the directory in which to save the filter's state
        :param checkpoint_prefix: the prefix of the checkpoint file to save
        :return: None
        """
        pass

    def restore_state_from_checkpoint(self, checkpoint_dir, checkpoint_prefix)->None:
        """
        Save the filter's internal state to a checkpoint to file, so that it can be later restored.
        :param checkpoint_dir: the directory from which to restore
        :param checkpoint_prefix: the checkpoint prefix to look for
        :return: None
        """
        pass

    def set_name(self, name: str) -> None:
        """
        Set the filter's name
        :param name: the filter's name
        :return: None
        """
        self.name = name


class OutputFilter(Filter):
    """
    An output filter is a module that filters the output from an agent to the environment.
    """
    def __init__(self, action_filters: OrderedDict([(str, 'ActionFilter')])=None,
                 is_a_reference_filter: bool=False, name=None):
        super().__init__(name)

        if action_filters is None:
            action_filters = OrderedDict([])
        self._action_filters = action_filters

        # We do not want to allow reference filters such as Atari to be used directly. These have to be duplicated first
        # and only then can change their values so to keep their original params intact for other agents in the graph.
        self.i_am_a_reference_filter = is_a_reference_filter

    def __call__(self):
        duplicate = deepcopy(self)
        duplicate.i_am_a_reference_filter = False
        return duplicate

    def set_device(self, device, memory_backend_params=None, mode='numpy') -> None:
        """
        An optional function that allows the filter to get the device if it is required to use tensorflow ops
        :param device: the device to use
        :return: None
        """
        [f.set_device(device, memory_backend_params, mode='numpy') for f in self.action_filters.values()]

    def set_session(self, sess) -> None:
        """
        An optional function that allows the filter to get the session if it is required to use tensorflow ops
        :param sess: the session
        :return: None
        """
        [f.set_session(sess) for f in self.action_filters.values()]

    def filter(self, action_info: ActionInfo) -> ActionInfo:
        """
        A wrapper around _filter which first copies the action_info so that we don't change the original one
        This function should not be updated!
        :param action_info: the input action_info
        :return: the filtered action_info
        """
        if self.i_am_a_reference_filter:
            raise Exception("The filter being used is a reference filter. It is not to be used directly. "
                            "Instead get a duplicate from it by calling __call__.")
        if len(self.action_filters.values()) == 0:
            return action_info
        filtered_action_info = copy.deepcopy(action_info)
        filtered_action = filtered_action_info.action
        for filter in reversed(self.action_filters.values()):
            filtered_action = filter.filter(filtered_action)

        filtered_action_info.action = filtered_action

        return filtered_action_info

    def reverse_filter(self, action_info: ActionInfo) -> ActionInfo:
        """
        A wrapper around _reverse_filter which first copies the action_info so that we don't change the original one
        This function should not be updated!
        :param action_info: the input action_info
        :return: the filtered action_info
        """
        if self.i_am_a_reference_filter:
            raise Exception("The filter being used is a reference filter. It is not to be used directly. "
                            "Instead get a duplicate from it by calling __call__.")
        filtered_action_info = copy.deepcopy(action_info)
        filtered_action = filtered_action_info.action
        for filter in self.action_filters.values():
            filter.validate_output_action(filtered_action)
            filtered_action = filter.reverse_filter(filtered_action)

        filtered_action_info.action = filtered_action

        return filtered_action_info

    def get_unfiltered_action_space(self, output_action_space: ActionSpace) -> ActionSpace:
        """
        Given the output action space, returns the corresponding unfiltered action space
        This function should not be updated!
        :param output_action_space: the output action space
        :return: the unfiltered action space
        """
        unfiltered_action_space = copy.deepcopy(output_action_space)
        for filter in self._action_filters.values():
            new_unfiltered_action_space = filter.get_unfiltered_action_space(unfiltered_action_space)
            filter.validate_output_action_space(unfiltered_action_space)
            unfiltered_action_space = new_unfiltered_action_space
        return unfiltered_action_space

    def reset(self) -> None:
        """
        Reset any internal memory stored in the filter.
        This function should not be updated!
        This is useful for stateful filters which stores information on previous filter calls.
        :return: None
        """
        [action_filter.reset() for action_filter in self._action_filters.values()]

    @property
    def action_filters(self) -> OrderedDict([(str, 'ActionFilter')]):
        return self._action_filters

    @action_filters.setter
    def action_filters(self, val: OrderedDict([(str, 'ActionFilter')])):
        self._action_filters = val

    def add_action_filter(self, filter_name: str, filter: 'ActionFilter', add_as_the_first_filter: bool=False):
        """
        Add an action filter to the filters list
        :param filter_name: the filter name
        :param filter: the filter to add
        :param add_as_the_first_filter: add the filter to the top of the filters stack
        :return: None
        """
        self._action_filters[filter_name] = filter
        if add_as_the_first_filter:
            self._action_filters.move_to_end(filter_name, last=False)

    def remove_action_filter(self, filter_name: str) -> None:
        """
        Remove an action filter from the filters list
        :param filter_name: the filter name
        :return: None
        """
        del self._action_filters[filter_name]

    def save_state_to_checkpoint(self, checkpoint_dir, checkpoint_prefix):
        """
        Currently not in use for OutputFilter.
        :param checkpoint_dir: the directory in which to save the filter's state
        :param checkpoint_prefix: the prefix of the checkpoint file to save
        :return:
        """
        pass

    def restore_state_from_checkpoint(self, checkpoint_dir, checkpoint_prefix)->None:
        """
        Currently not in use for OutputFilter.
        :param checkpoint_dir: the directory from which to restore
        :param checkpoint_prefix: the checkpoint prefix to look for
        :return: None
        """
        pass



class NoOutputFilter(OutputFilter):
    """
    Creates an empty output filter. Used only for readability when creating the presets
    """
    def __init__(self):
        super().__init__(is_a_reference_filter=False)


class InputFilter(Filter):
    """
    An input filter is a module that filters the input from an environment to the agent.
    """
    def __init__(self, observation_filters: Dict[str, Dict[str, 'ObservationFilter']]=None,
                 reward_filters: Dict[str, 'RewardFilter']=None,
                 is_a_reference_filter: bool=False, name=None):
        super().__init__(name)
        if observation_filters is None:
            observation_filters = {}
        if reward_filters is None:
            reward_filters = OrderedDict([])
        self._observation_filters = observation_filters
        self._reward_filters = reward_filters

        # We do not want to allow reference filters such as Atari to be used directly. These have to be duplicated first
        # and only then can change their values so to keep their original params intact for other agents in the graph.
        self.i_am_a_reference_filter = is_a_reference_filter

    def __call__(self):
        duplicate = deepcopy(self)
        duplicate.i_am_a_reference_filter = False
        return duplicate

    def set_device(self, device, memory_backend_params=None, mode='numpy') -> None:
        """
        An optional function that allows the filter to get the device if it is required to use tensorflow ops
        :param device: the device to use
        :return: None
        """
        [f.set_device(device, memory_backend_params, mode) for f in self.reward_filters.values()]
        [[f.set_device(device, memory_backend_params, mode) for f in filters.values()] for filters in self.observation_filters.values()]

    def set_session(self, sess) -> None:
        """
        An optional function that allows the filter to get the session if it is required to use tensorflow ops
        :param sess: the session
        :return: None
        """
        [f.set_session(sess) for f in self.reward_filters.values()]
        [[f.set_session(sess) for f in filters.values()] for filters in self.observation_filters.values()]

    def filter(self, unfiltered_data: Union[EnvResponse, List[EnvResponse], Transition, List[Transition]],
               update_internal_state: bool=True, deep_copy: bool=True) -> Union[List[EnvResponse], List[Transition]]:
        """
        A wrapper around _filter which first copies the env_response so that we don't change the original one
        This function should not be updated!
        :param unfiltered_data: the input data
        :param update_internal_state: should the filter's internal state change due to this call
        :return: the filtered env_response
        """
        if self.i_am_a_reference_filter:
            raise Exception("The filter being used is a reference filter. It is not to be used directly. "
                            "Instead get a duplicate from it by calling __call__.")
        if deep_copy:
            filtered_data = copy.deepcopy(unfiltered_data)
        else:
            filtered_data = [copy.copy(t) for t in unfiltered_data]
        filtered_data = force_list(filtered_data)

        # TODO: implement observation space validation
        # filter observations
        if isinstance(filtered_data[0], Transition):
            state_objects_to_filter = [[f.state for f in filtered_data],
                                       [f.next_state for f in filtered_data]]
        elif isinstance(filtered_data[0], EnvResponse):
            state_objects_to_filter = [[f.next_state for f in filtered_data]]
        else:
            raise ValueError("unfiltered_data should be either of type EnvResponse or Transition. ")

        for state_object_list in state_objects_to_filter:
            for observation_name, filters in self._observation_filters.items():
                if observation_name in state_object_list[0].keys():
                    for filter in filters.values():
                        data_to_filter = [state_object[observation_name] for state_object in state_object_list]
                        if filter.supports_batching:
                            filtered_observations = filter.filter(
                                data_to_filter, update_internal_state=update_internal_state)
                        else:
                            filtered_observations = []
                            for data_point in data_to_filter:
                                filtered_observations.append(filter.filter(
                                    data_point, update_internal_state=update_internal_state))

                        for i, state_object in enumerate(state_object_list):
                            state_object[observation_name] = filtered_observations[i]

        # filter reward
        for f in filtered_data:
            filtered_reward = f.reward
            for filter in self._reward_filters.values():
                filtered_reward = filter.filter(filtered_reward, update_internal_state)
            f.reward = filtered_reward

        return filtered_data

    def get_filtered_observation_space(self, observation_name: str,
                                       input_observation_space: ObservationSpace) -> ObservationSpace:
        """
        Given the input observation space, returns the corresponding filtered observation space
        This function should not be updated!
        :param observation_name: the name of the observation to which we want to calculate the filtered space
        :param input_observation_space: the input observation space
        :return: the filtered observation space
        """
        filtered_observation_space = copy.deepcopy(input_observation_space)
        if observation_name in self._observation_filters.keys():
            for filter in self._observation_filters[observation_name].values():
                filter.validate_input_observation_space(filtered_observation_space)
                filtered_observation_space = filter.get_filtered_observation_space(filtered_observation_space)
        return filtered_observation_space

    def get_filtered_reward_space(self, input_reward_space: RewardSpace) -> RewardSpace:
        """
        Given the input reward space, returns the corresponding filtered reward space
        This function should not be updated!
        :param input_reward_space: the input reward space
        :return: the filtered reward space
        """
        filtered_reward_space = copy.deepcopy(input_reward_space)
        for filter in self._reward_filters.values():
            filtered_reward_space = filter.get_filtered_reward_space(filtered_reward_space)
        return filtered_reward_space

    def reset(self) -> None:
        """
        Reset any internal memory stored in the filter.
        This function should not be updated!
        This is useful for stateful filters which stores information on previous filter calls.
        :return: None
        """
        for curr_observation_filters in self._observation_filters.values():
            [observation_filter.reset() for observation_filter in curr_observation_filters.values()]
        [reward_filter.reset() for reward_filter in self._reward_filters.values()]

    @property
    def observation_filters(self) -> Dict[str, Dict[str, 'ObservationFilter']]:
        return self._observation_filters

    @observation_filters.setter
    def observation_filters(self, val: Dict[str, Dict[str, 'ObservationFilter']]):
        self._observation_filters = val

    @property
    def reward_filters(self) -> OrderedDict([(str, 'RewardFilter')]):
        return self._reward_filters

    @reward_filters.setter
    def reward_filters(self, val: OrderedDict([(str, 'RewardFilter')])):
        self._reward_filters = val

    def copy_filters_from_one_observation_to_another(self, from_observation: str, to_observation: str):
        """
        Copy all the filters created for some observation to another observation
        :param from_observation: the source observation to copy from
        :param to_observation: the target observation to copy to
        :return: None
        """
        self._observation_filters[to_observation] = copy.deepcopy(self._observation_filters[from_observation])

    def add_observation_filter(self, observation_name: str, filter_name: str, filter: 'ObservationFilter',
                               add_as_the_first_filter: bool=False):
        """
        Add an observation filter to the filters list
        :param observation_name: the name of the observation to apply to
        :param filter_name: the filter name
        :param filter: the filter to add
        :param add_as_the_first_filter: add the filter to the top of the filters stack
        :return: None
        """
        if observation_name not in self._observation_filters.keys():
            self._observation_filters[observation_name] = OrderedDict()
        self._observation_filters[observation_name][filter_name] = filter
        if add_as_the_first_filter:
            self._observation_filters[observation_name].move_to_end(filter_name, last=False)

    def add_reward_filter(self, filter_name: str, filter: 'RewardFilter', add_as_the_first_filter: bool=False):
        """
        Add a reward filter to the filters list
        :param filter_name: the filter name
        :param filter: the filter to add
        :param add_as_the_first_filter: add the filter to the top of the filters stack
        :return: None
        """
        self._reward_filters[filter_name] = filter
        if add_as_the_first_filter:
            self._reward_filters.move_to_end(filter_name, last=False)

    def remove_observation_filter(self, observation_name: str, filter_name: str) -> None:
        """
        Remove an observation filter from the filters list
        :param observation_name: the name of the observation to apply to
        :param filter_name: the filter name
        :return: None
        """
        del self._observation_filters[observation_name][filter_name]

    def remove_reward_filter(self, filter_name: str) -> None:
        """
        Remove a reward filter from the filters list
        :param filter_name: the filter name
        :return: None
        """
        del self._reward_filters[filter_name]

    def save_state_to_checkpoint(self, checkpoint_dir, checkpoint_prefix):
        """
        Save the filter's internal state to a checkpoint to file, so that it can be later restored.
        :param checkpoint_dir: the directory in which to save the filter's state
        :param checkpoint_prefix: the prefix of the checkpoint file to save
        :return: None
        """
        checkpoint_prefix = '.'.join([checkpoint_prefix, 'filters'])
        if self.name is not None:
            checkpoint_prefix = '.'.join([checkpoint_prefix, self.name])
        for filter_name, filter in self._reward_filters.items():
            curr_reward_filter_ckpt_prefix = '.'.join([checkpoint_prefix, 'reward_filters', filter_name])
            filter.save_state_to_checkpoint(checkpoint_dir, curr_reward_filter_ckpt_prefix)

        for observation_name, filters_dict in self._observation_filters.items():
            for filter_name, filter in filters_dict.items():
                curr_obs_filter_ckpt_prefix = '.'.join([checkpoint_prefix, 'observation_filters', observation_name,
                                                                 filter_name])
                filter.save_state_to_checkpoint(checkpoint_dir, curr_obs_filter_ckpt_prefix)

    def restore_state_from_checkpoint(self, checkpoint_dir, checkpoint_prefix)->None:
        """
        Save the filter's internal state to a checkpoint to file, so that it can be later restored.
        :param checkpoint_dir: the directory from which to restore
        :param checkpoint_prefix: the checkpoint prefix to look for
        :return: None
        """
        checkpoint_prefix = '.'.join([checkpoint_prefix, 'filters'])
        if self.name is not None:
            checkpoint_prefix = '.'.join([checkpoint_prefix, self.name])
        for filter_name, filter in self._reward_filters.items():
            curr_reward_filter_ckpt_prefix = '.'.join([checkpoint_prefix, 'reward_filters', filter_name])
            filter.restore_state_from_checkpoint(checkpoint_dir, curr_reward_filter_ckpt_prefix)

        for observation_name, filters_dict in self._observation_filters.items():
            for filter_name, filter in filters_dict.items():
                curr_obs_filter_ckpt_prefix = '.'.join([checkpoint_prefix, 'observation_filters', observation_name,
                                                                 filter_name])
                filter.restore_state_from_checkpoint(checkpoint_dir, curr_obs_filter_ckpt_prefix)


class NoInputFilter(InputFilter):
    """
    Creates an empty input filter. Used only for readability when creating the presets
    """
    def __init__(self):
        super().__init__(is_a_reference_filter=False, name='no_input_filter')


